"""Representation of the contents of a cluster.

A manifest may be built directly from the local context of a cluster, or may be
serialized and stored and checked into the cluster for use in other applications
e.g. such as writing management plan for resources.
"""

import base64
from dataclasses import dataclass, field
import logging
from pathlib import Path
from typing import Any, Optional, cast

import aiofiles
from mashumaro.codecs.yaml import yaml_decode, yaml_encode
from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import BaseConfig

from .exceptions import InputException

__all__ = [
    "read_manifest",
    "write_manifest",
    "Manifest",
    "Cluster",
    "Kustomization",
    "HelmRepository",
    "HelmRelease",
    "HelmChart",
    "ConfigMap",
    "Secret",
]

_LOGGER = logging.getLogger(__name__)


# Match a prefix of apiVersion to ensure we have the right type of object.
# We don't check specific versions for forward compatibility on upgrade.
FLUXTOMIZE_DOMAIN = "kustomize.toolkit.fluxcd.io"
KUSTOMIZE_DOMAIN = "kustomize.config.k8s.io"
HELM_REPO_DOMAIN = "source.toolkit.fluxcd.io"
HELM_RELEASE_DOMAIN = "helm.toolkit.fluxcd.io"
OCI_REPOSITORY_DOMAIN = "source.toolkit.fluxcd.io"
CRD_KIND = "CustomResourceDefinition"
SECRET_KIND = "Secret"
CONFIG_MAP_KIND = "ConfigMap"
DEFAULT_NAMESPACE = "flux-system"
VALUE_PLACEHOLDER_TEMPLATE = "..PLACEHOLDER_{name}.."
HELM_RELEASE = "HelmRelease"
HELM_REPOSITORY = "HelmRepository"
GIT_REPOSITORY = "GitRepository"
OCI_REPOSITORY = "OCIRepository"


REPO_TYPE_DEFAULT = "default"
REPO_TYPE_OCI = "oci"


def _check_version(doc: dict[str, Any], version: str) -> None:
    """Assert that the resource has the specified version."""
    if not (api_version := doc.get("apiVersion")):
        raise InputException(f"Invalid object missing apiVersion: {doc}")
    if not api_version.startswith(version):
        raise InputException(f"Invalid object expected '{version}': {doc}")


@dataclass
class BaseManifest(DataClassDictMixin):
    """Base class for all manifest objects."""

    def compact_dict(self) -> dict[str, Any]:
        """Return a compact dictionary representation of the object.

        This is similar to `dict()` but with a specific implementation for serializing
        with variable fields removed.
        """
        return self.to_dict()

    @classmethod
    def parse_yaml(cls, content: str) -> "BaseManifest":
        """Parse a serialized manifest."""
        return yaml_decode(content, cls)

    def yaml(self, exclude: dict[str, Any] | None = None) -> str:
        """Return a YAML string representation of compact_dict."""
        return yaml_encode(self, self.__class__)  # type: ignore[return-value]

    class Config(BaseConfig):
        omit_none = True


@dataclass(frozen=True, order=True)
class NamedResource:
    """Identifier for a kubernetes resource."""

    kind: str
    namespace: str | None
    name: str

    @property
    def namespaced_name(self) -> str:
        if self.namespace:
            return f"{self.namespace}/{self.name}"
        return self.name


@dataclass
class HelmChart(BaseManifest):
    """A representation of an instantiation of a chart for a HelmRelease."""

    name: str
    """The name of the chart within the HelmRepository."""

    version: Optional[str] = field(metadata={"serialize": "omit"})
    """The version of the chart."""

    repo_name: str
    """The short name of the repository."""

    repo_namespace: str
    """The namespace of the repository."""

    repo_kind: str = HELM_REPOSITORY
    """The kind of the soruceRef of the repository (e.g. HelmRepository, GitRepository)."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any], default_namespace: str) -> "HelmChart":
        """Parse a HelmChart from a HelmRelease resource object."""
        _check_version(doc, HELM_RELEASE_DOMAIN)
        if not (spec := doc.get("spec")):
            raise InputException(f"Invalid {cls} missing spec: {doc}")
        chart_ref = spec.get("chartRef")
        chart = spec.get("chart")
        if not chart_ref and not chart:
            raise InputException(
                f"Invalid {cls} missing spec.chart or spec.chartRef: {doc}"
            )
        if chart_ref:
            if not (kind := chart_ref.get("kind")):
                raise InputException(f"Invalid {cls} missing spec.chartRef.kind: {doc}")
            if not (name := chart_ref.get("name")):
                raise InputException(f"Invalid {cls} missing spec.chartRef.name: {doc}")
            if not (namespace := chart_ref.get("namespace")):
                raise InputException(
                    f"Invalid {cls} missing spec.chartRef.namespace: {doc}"
                )
            return cls(
                name=name,
                version=None,
                repo_name=name,
                repo_namespace=namespace,
                repo_kind=kind,
            )
        if not (chart_spec := chart.get("spec")):
            raise InputException(f"Invalid {cls} missing spec.chart.spec: {doc}")
        if not (chart := chart_spec.get("chart")):
            raise InputException(f"Invalid {cls} missing spec.chart.spec.chart: {doc}")
        version = chart_spec.get("version")
        if not (source_ref := chart_spec.get("sourceRef")):
            raise InputException(
                f"Invalid {cls} missing spec.chart.spec.sourceRef: {doc}"
            )
        if "name" not in source_ref:
            raise InputException(f"Invalid {cls} missing sourceRef fields: {doc}")
        return cls(
            name=chart,
            version=version,
            repo_name=source_ref["name"],
            repo_namespace=source_ref.get("namespace", default_namespace),
            repo_kind=source_ref.get("kind", HELM_REPOSITORY),
        )

    @property
    def repo_full_name(self) -> str:
        """Identifier for the HelmRepository."""
        return f"{self.repo_namespace}-{self.repo_name}"

    @property
    def chart_name(self) -> str:
        """Identifier for the HelmChart."""
        return f"{self.repo_full_name}/{self.name}"


@dataclass
class ValuesReference(BaseManifest):
    """A reference to a resource containing values for a HelmRelease."""

    kind: str
    """The kind of resource."""

    name: str
    """The name of the resource."""

    values_key: str = field(
        metadata=field_options(alias="valuesKey"), default="values.yaml"
    )
    """The key in the resource that contains the values."""

    target_path: Optional[str] = field(
        metadata=field_options(alias="targetPath"), default=None
    )
    """The path in the HelmRelease values to store the values."""

    optional: bool = False
    """Whether the reference is optional."""


@dataclass
class HelmRelease(BaseManifest):
    """A representation of a Flux HelmRelease."""

    name: str
    """The name of the HelmRelease."""

    namespace: str
    """The namespace that owns the HelmRelease."""

    chart: HelmChart
    """A mapping to a specific helm chart for this HelmRelease."""

    values: Optional[dict[str, Any]] = field(
        metadata={"serialize": "omit"}, default=None
    )
    """The values to install in the chart."""

    values_from: Optional[list[ValuesReference]] = field(
        metadata={"serialize": "omit"}, default=None
    )
    """A list of values to reference from an ConfigMap or Secret."""

    images: list[str] | None = field(default=None)
    """The list of images referenced in the HelmRelease."""

    labels: dict[str, str] | None = field(metadata={"serialize": "omit"}, default=None)
    """A list of labels on the HelmRelease."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "HelmRelease":
        """Parse a HelmRelease from a kubernetes resource object."""
        _check_version(doc, HELM_RELEASE_DOMAIN)
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise InputException(f"Invalid {cls} missing metadata.namespace: {doc}")
        chart = HelmChart.parse_doc(doc, namespace)
        spec = doc["spec"]
        values_from: list[ValuesReference] | None = None
        if values_from_dict := spec.get("valuesFrom"):
            values_from = [
                ValuesReference.from_dict(subdoc) for subdoc in values_from_dict
            ]
        return HelmRelease(
            name=name,
            namespace=namespace,
            chart=chart,
            values=spec.get("values"),
            values_from=values_from,
            labels=metadata.get("labels"),
        )

    @property
    def release_name(self) -> str:
        """Identifier for the HelmRelease."""
        return f"{self.namespace}-{self.name}"

    @property
    def repo_name(self) -> str:
        """Identifier for the HelmRepository identified in the HelmChart."""
        return f"{self.chart.repo_namespace}-{self.chart.repo_name}"

    @property
    def namespaced_name(self) -> str:
        """Return the namespace and name concatenated as an id."""
        return f"{self.namespace}/{self.name}"

    @property
    def resource_dependencies(self) -> list[NamedResource]:
        """Return the list of input dependencies for the HelmRelease."""
        deps = [
            NamedResource(
                kind=HELM_RELEASE,
                name=self.name,
                namespace=self.namespace,
            )
        ]
        if self.chart:
            deps.append(
                NamedResource(
                    kind=self.chart.repo_kind,
                    name=self.chart.repo_name,
                    namespace=self.chart.repo_namespace,
                )
            )
        names_seen = set()
        for ref in self.values_from or ():
            if ref.name in names_seen:
                continue
            names_seen.add(ref.name)
            deps.append(
                NamedResource(kind=ref.kind, name=ref.name, namespace=self.namespace)
            )
        return deps


@dataclass
class HelmRepository(BaseManifest):
    """A representation of a flux HelmRepository."""

    name: str
    """The name of the HelmRepository."""

    namespace: str
    """The namespace of owning the HelmRepository."""

    url: str
    """The URL to the repository of helm charts."""

    repo_type: str | None = None
    """The type of the HelmRepository."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "HelmRepository":
        """Parse a HelmRepository from a kubernetes resource."""
        _check_version(doc, HELM_REPO_DOMAIN)
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise InputException(f"Invalid {cls} missing metadata.namespace: {doc}")
        if not (spec := doc.get("spec")):
            raise InputException(f"Invalid {cls} missing spec: {doc}")
        if not (url := spec.get("url")):
            raise InputException(f"Invalid {cls} missing spec.url: {doc}")
        return cls(
            name=name,
            namespace=namespace,
            url=url,
            repo_type=spec.get("type", REPO_TYPE_DEFAULT),
        )

    @property
    def repo_name(self) -> str:
        """Identifier for the HelmRepository."""
        return f"{self.namespace}-{self.name}"


@dataclass
class OCIRepository(BaseManifest):
    """A representation of a flux OCIRepository."""

    name: str
    """The name of the OCIRepository."""

    namespace: str
    """The namespace of owning the OCIRepository."""

    url: str
    """The URL to the repository."""

    ref_tag: str | None = None
    """The version tag of the repository."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "OCIRepository":
        """Parse a HelmRepository from a kubernetes resource."""
        _check_version(doc, OCI_REPOSITORY_DOMAIN)
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise InputException(f"Invalid {cls} missing metadata.namespace: {doc}")
        if not (spec := doc.get("spec")):
            raise InputException(f"Invalid {cls} missing spec: {doc}")
        if not (url := spec.get("url")):
            raise InputException(f"Invalid {cls} missing spec.url: {doc}")
        ref_tag = spec.get("ref", {}).get("tag")
        return cls(
            name=name,
            namespace=namespace,
            url=url,
            ref_tag=ref_tag,
        )

    @property
    def repo_name(self) -> str:
        """Identifier for the OCIRepository."""
        return f"{self.namespace}-{self.name}"


@dataclass
class ConfigMap(BaseManifest):
    """A ConfigMap is an API object used to store data in key-value pairs."""

    name: str
    """The name of the kustomization."""

    namespace: str | None = None
    """The namespace of the kustomization."""

    data: dict[str, Any] | None = field(metadata={"serialize": "omit"}, default=None)
    """The data in the ConfigMap."""

    binary_data: dict[str, Any] | None = field(
        metadata={"serialize": "omit"}, default=None
    )
    """The binary data in the ConfigMap."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "ConfigMap":
        """Parse a config map object from a kubernetes resource."""
        _check_version(doc, "v1")
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        namespace = metadata.get("namespace")
        return ConfigMap(
            name=name,
            namespace=namespace,
            data=doc.get("data"),
            binary_data=doc.get("binaryData"),
        )


@dataclass
class Secret(BaseManifest):
    """A Secret contains a small amount of sensitive data."""

    name: str
    """The name of the kustomization."""

    namespace: str | None = None
    """The namespace of the kustomization."""

    data: dict[str, Any] | None = field(metadata={"serialize": "omit"}, default=None)
    """The data in the Secret."""

    string_data: dict[str, Any] | None = field(
        metadata={"serialize": "omit"}, default=None
    )
    """The string data in the Secret."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "Secret":
        """Parse a secret object from a kubernetes resource."""
        _check_version(doc, "v1")
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        namespace = metadata.get("namespace")
        # While secrets are not typically stored in the cluster, we replace with
        # placeholder values anyway.
        if data := doc.get("data"):
            for key, value in data.items():
                data[key] = base64.b64encode(
                    VALUE_PLACEHOLDER_TEMPLATE.format(name=key).encode()
                )
        if string_data := doc.get("stringData"):
            for key, value in string_data.items():
                string_data[key] = VALUE_PLACEHOLDER_TEMPLATE.format(name=key)
        return Secret(
            name=name, namespace=namespace, data=data, string_data=string_data
        )


@dataclass
class SubstituteReference(BaseManifest):
    """SubstituteReference contains a reference to a resource containing the variables name and value."""

    kind: str
    """The kind of resource."""

    name: str
    """The name of the resource."""

    optional: bool = False
    """Whether the reference is optional."""


@dataclass
class Kustomization(BaseManifest):
    """A Kustomization is a set of declared cluster artifacts.

    This represents a flux Kustomization that points to a path that
    contains typical `kustomize` Kustomizations on local disk that
    may be flat or contain overlays.
    """

    name: str
    """The name of the kustomization."""

    namespace: str | None
    """The namespace of the kustomization."""

    path: str
    """The local repo path to the kustomization contents."""

    helm_repos: list[HelmRepository] = field(default_factory=list)
    """The set of HelmRepositories represented in this kustomization."""

    oci_repos: list[OCIRepository] = field(default_factory=list)
    """The set of OCIRepositories represented in this kustomization."""

    helm_releases: list[HelmRelease] = field(default_factory=list)
    """The set of HelmRelease represented in this kustomization."""

    config_maps: list[ConfigMap] = field(default_factory=list)
    """The list of config maps referenced in the kustomization."""

    secrets: list[Secret] = field(default_factory=list)
    """The list of secrets referenced in the kustomization."""

    source_path: str | None = field(metadata={"serialize": "omit"}, default=None)
    """Optional source path for this Kustomization, relative to the build path."""

    source_kind: str | None = field(metadata={"serialize": "omit"}, default=None)
    """The sourceRef kind that provides this Kustomization e.g. GitRepository etc."""

    source_name: str | None = field(metadata={"serialize": "omit"}, default=None)
    """The name of the sourceRef that provides this Kustomization."""

    source_namespace: str | None = field(metadata={"serialize": "omit"}, default=None)
    """The namespace of the sourceRef that provides this Kustomization."""

    target_namespace: str | None = field(metadata={"serialize": "omit"}, default=None)
    """The namespace to target when performing the operation."""

    contents: dict[str, Any] | None = field(
        metadata={"serialize": "omit"}, default=None
    )
    """Contents of the raw Kustomization document."""

    images: list[str] | None = field(default=None)
    """The list of images referenced in the kustomization."""

    postbuild_substitute: Optional[dict[str, Any]] = field(
        metadata={"serialize": "omit"}, default=None
    )
    """A map of key/value pairs to substitute into the final YAML manifest, after building."""

    postbuild_substitute_from: Optional[list[SubstituteReference]] = field(
        metadata={"serialize": "omit"}, default=None
    )
    """A list of substitutions to reference from an ConfigMap or Secret."""

    depends_on: list[str] | None = field(metadata={"serialize": "omit"}, default=None)
    """A list of namespaced names that this Kustomization depends on."""

    labels: dict[str, str] | None = field(metadata={"serialize": "omit"}, default=None)
    """A list of labels on the Kustomization."""

    @classmethod
    def parse_doc(cls, doc: dict[str, Any]) -> "Kustomization":
        """Parse a partial Kustomization from a kubernetes resource."""
        _check_version(doc, FLUXTOMIZE_DOMAIN)
        if not (metadata := doc.get("metadata")):
            raise InputException(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise InputException(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise InputException(f"Invalid {cls} missing metadata.namespace: {doc}")
        if not (spec := doc.get("spec")):
            raise InputException(f"Invalid {cls} missing spec: {doc}")
        path = spec.get("path", "")
        source_path = metadata.get("annotations", {}).get("config.kubernetes.io/path")
        source_ref = spec.get("sourceRef", {})
        postbuild = spec.get("postBuild", {})
        substitute_from: list[SubstituteReference] | None = None
        if substitute_from_dict := postbuild.get("substituteFrom"):
            substitute_from = [
                SubstituteReference(**subdoc) for subdoc in substitute_from_dict
            ]
        depends_on = []
        for dependency in spec.get("dependsOn", ()):
            if not (dep_name := dependency.get("name")):
                raise InputException(f"Invalid {cls} missing dependsOn.name: {doc}")
            dep_namespace = dependency.get("namespace", namespace)
            depends_on.append(f"{dep_namespace}/{dep_name}")
        return Kustomization(
            name=name,
            namespace=namespace,
            path=path,
            source_path=source_path,
            source_kind=source_ref.get("kind"),
            source_name=source_ref.get("name"),
            source_namespace=source_ref.get("namespace", namespace),
            target_namespace=spec.get("targetNamespace"),
            contents=doc,
            postbuild_substitute=postbuild.get("substitute"),
            postbuild_substitute_from=substitute_from,
            depends_on=depends_on,
            labels=metadata.get("labels"),
        )

    @property
    def id_name(self) -> str:
        """Identifier for the Kustomization in tests"""
        return f"{self.path}"

    @property
    def namespaced_name(self) -> str:
        """Return the namespace and name concatenated as an id."""
        return f"{self.namespace}/{self.name}"

    def validate_depends_on(self, all_ks: set[str]) -> None:
        """Validate depends_on values are all correct given the list of Kustomizations."""
        depends_on = set(self.depends_on or {})
        if missing := (depends_on - all_ks):
            _LOGGER.warning(
                "Kustomization %s has dependsOn with invalid names: %s",
                self.namespaced_name,
                missing,
            )
            self.depends_on = list(depends_on - missing)

    def update_postbuild_substitutions(self, substitutions: dict[str, Any]) -> None:
        """Update the postBuild.substitutions in the extracted values and raw doc contents."""
        if self.postbuild_substitute is None:
            self.postbuild_substitute = {}
        self.postbuild_substitute.update(substitutions)
        if self.contents:
            post_build = self.contents["spec"]["postBuild"]
            if (substitute := post_build.get("substitute")) is None:
                substitute = {}
                post_build["substitute"] = substitute
            substitute.update(substitutions)


@dataclass
class Cluster(BaseManifest):
    """A set of nodes that run containerized applications.

    Many flux git repos will only have a single flux cluster, though
    a repo may also contain multiple (e.g. dev an prod).
    """

    path: str
    """The local git repo path to the Kustomization objects for the cluster."""

    kustomizations: list[Kustomization] = field(default_factory=list)
    """A list of flux Kustomizations for the cluster."""

    @property
    def id_name(self) -> str:
        """Identifier for the Cluster in tests."""
        return f"{self.path}"

    @property
    def helm_repos(self) -> list[HelmRepository]:
        """Return the list of HelmRepository objects from all Kustomizations."""
        return [
            repo
            for kustomization in self.kustomizations
            for repo in kustomization.helm_repos
        ]

    @property
    def oci_repos(self) -> list[OCIRepository]:
        """Return the list of OCIRepository objects from all Kustomizations."""
        return [
            repo
            for kustomization in self.kustomizations
            for repo in kustomization.oci_repos
        ]

    @property
    def helm_releases(self) -> list[HelmRelease]:
        """Return the list of HelmRelease objects from all Kustomizations."""
        return [
            release
            for kustomization in self.kustomizations
            for release in kustomization.helm_releases
        ]


@dataclass
class Manifest(BaseManifest):
    """Holds information about cluster and applications contained in a repo."""

    clusters: list[Cluster]
    """A list of Clusters represented in the repo."""


async def read_manifest(manifest_path: Path) -> Manifest:
    """Return the contents of a serialized manifest file.

    A manifest file is typically created by `flux-local get cluster -o yaml` or
    similar command.
    """
    async with aiofiles.open(str(manifest_path)) as manifest_file:
        content = await manifest_file.read()
        if not content:
            raise ValueError("validation error for Manifest file {manifest_path}")
        return cast(Manifest, Manifest.parse_yaml(content))


async def write_manifest(manifest_path: Path, manifest: Manifest) -> None:
    """Write the specified manifest content to disk."""
    content = manifest.yaml()
    async with aiofiles.open(str(manifest_path), mode="w") as manifest_file:
        await manifest_file.write(content)


async def update_manifest(manifest_path: Path, manifest: Manifest) -> None:
    """Write the specified manifest only if changed."""
    new_content = manifest.yaml()
    async with aiofiles.open(str(manifest_path)) as manifest_file:
        content = await manifest_file.read()
    if content == new_content:
        return
    async with aiofiles.open(str(manifest_path), mode="w") as manifest_file:
        await manifest_file.write(new_content)



"""Flux-local diff action."""

import asyncio
import functools
import os
from argparse import ArgumentParser, _SubParsersAction as SubParsersAction
from contextlib import contextmanager
import logging
import pathlib
import shlex
import tempfile
from typing import cast, Generator, Any


from flux_local import git_repo
from flux_local.visitor import HelmVisitor, ObjectOutput
from flux_local.resource_diff import (
    get_helm_release_diff_keys,
    perform_yaml_diff,
    perform_external_diff,
    perform_object_diff,
    build_helm_dependency_map,
    perform_json_diff,
)

from . import selector

_LOGGER = logging.getLogger(__name__)

# Type for command line flags of comma separated list
_CSV = functools.partial(str.split, sep=",")


def add_diff_flags(args: ArgumentParser) -> None:
    """Add shared diff flags."""
    args.add_argument(
        "--output",
        "-o",
        choices=["diff", "yaml", "object", "json"],
        default="diff",
        help="Output format of the command",
    )
    args.add_argument(
        "--unified",
        "-u",
        type=int,
        default=3,
        help="output NUM (default 3) lines of unified context",
    )
    args.add_argument(
        "--path-orig",
        help="Path to compare against, or creates a work tree if not specified",
        type=pathlib.Path,
        default=None,
        nargs="?",
    )

    args.add_argument(
        "--branch-orig",
        help="Branch to compare against using worktree",
        type=str,
        default=None,
        nargs="?",
    )

    args.add_argument(
        "--strip-attrs",
        help="Labels or annotations to strip from the diff",
        type=_CSV,
    )
    args.add_argument(
        "--limit-bytes",
        help="Maximum bytes for each diff output (0=unlimited)",
        type=int,
        default=0,
    )


@contextmanager
def create_diff_path(
    selector: git_repo.PathSelector,
    **kwargs: Any,
) -> Generator[git_repo.PathSelector, None, None]:
    """Create a context manager for the diff path.

    This will create a new worktree by default, or use the path in the flags
    which is useful when run from CI.
    """
    if path_orig := kwargs.get("path_orig"):
        yield git_repo.PathSelector(path_orig, sources=kwargs.get("sources"))
        return

    with git_repo.create_worktree(
        selector.repo, existing_branch=kwargs.get("branch_orig")
    ) as worktree:
        yield git_repo.PathSelector(pathlib.Path(worktree) / selector.relative_path)


class DiffKustomizationAction:
    """Flux-local diff for Kustomizations."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args: ArgumentParser = cast(
            ArgumentParser,
            subparsers.add_parser(
                "kustomizations",
                aliases=["ks", "kustomization"],
                help="Diff Kustomization objects",
                description=(
                    "The diff command does a local kustomize build compared "
                    "with the repo and prints the diff."
                ),
            ),
        )
        args.add_argument(
            "--output-file",
            type=str,
            default="/dev/stdout",
            help="Output file for the results of the command",
        )
        selector.add_ks_selector_flags(args)
        add_diff_flags(args)
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        output: str,
        unified: int,
        strip_attrs: list[str] | None,
        limit_bytes: int,
        output_file: str,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        query = selector.build_ks_selector(**kwargs)
        query.helm_release.enabled = False

        content = ObjectOutput(strip_attrs)
        query.kustomization.visitor = content.visitor()
        await git_repo.build_manifest(
            selector=query, options=selector.options(**kwargs)
        )

        orig_content = ObjectOutput(strip_attrs)
        with create_diff_path(query.path, **kwargs) as path_selector:
            query.path = path_selector
            query.kustomization.visitor = orig_content.visitor()
            await git_repo.build_manifest(
                selector=query, options=selector.options(**kwargs)
            )

        if not orig_content.content and not content.content:
            print(selector.not_found("Kustomization", query.kustomization))
            return

        _LOGGER.debug("Diffing content")
        with open(output_file, "w") as file:
            if output == "yaml":
                result = perform_yaml_diff(orig_content, content, unified, limit_bytes)
                for line in result:
                    print(line, file=file)
            elif output == "json":
                result = perform_json_diff(orig_content, content, unified, limit_bytes)
                for line in result:
                    print(line, file=file)
            elif external_diff := os.environ.get("DIFF"):
                async for line in perform_external_diff(
                    shlex.split(external_diff), orig_content, content, limit_bytes
                ):
                    print(line, file=file)
            else:
                result = perform_object_diff(
                    orig_content, content, unified, limit_bytes
                )
                for line in result:
                    print(line, file=file)


class DiffHelmReleaseAction:
    """Flux-local diff for HelmRelease."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args: ArgumentParser = cast(
            ArgumentParser,
            subparsers.add_parser(
                "helmreleases",
                aliases=["hr", "helmrelease"],
                help="Diff HelmRelease objects",
                description=(
                    "The diff command does a local kustomize build, then inflates "
                    "the helm template and compares with the repo and prints the diff."
                ),
            ),
        )
        args.add_argument(
            "--output-file",
            type=str,
            default="/dev/stdout",
            help="Output file for the results of the command",
        )
        selector.add_hr_selector_flags(args)
        selector.add_helm_options_flags(args)
        add_diff_flags(args)
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        output: str,
        unified: int,
        strip_attrs: list[str] | None,
        limit_bytes: int,
        output_file: str,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        query = selector.build_hr_selector(**kwargs)
        content = ObjectOutput(strip_attrs)
        helm_visitor = HelmVisitor()
        query.kustomization.visitor = content.visitor()
        query.helm_repo.visitor = helm_visitor.repo_visitor()
        query.oci_repo.visitor = helm_visitor.repo_visitor()
        query.helm_release.visitor = helm_visitor.release_visitor()
        options = selector.build_helm_options(**kwargs)
        await git_repo.build_manifest(
            selector=query, options=selector.options(**kwargs)
        )

        orig_content = ObjectOutput(strip_attrs)
        orig_helm_visitor = HelmVisitor()
        with create_diff_path(query.path, **kwargs) as path_selector:
            query.path = path_selector
            query.kustomization.visitor = orig_content.visitor()
            query.helm_repo.visitor = orig_helm_visitor.repo_visitor()
            query.oci_repo.visitor = orig_helm_visitor.repo_visitor()
            query.helm_release.visitor = orig_helm_visitor.release_visitor()
            await git_repo.build_manifest(
                selector=query, options=selector.options(**kwargs)
            )

        if not helm_visitor.releases and not orig_helm_visitor.releases:
            with open(output_file, "w") as file:
                print(selector.not_found("HelmRelease", query.helm_release), file=file)
            return

        # Find HelmRelease objects with diffs and prune all other HelmReleases from
        # the helm visitors. We assume that the only way for a HelmRelease output
        # to have a diff is if any of the resources that the HelmRelease
        # depends on in the kustomization has a diff. This avoids needing to
        # template every possible helm chart when nothing as changed.
        dependency_map = build_helm_dependency_map(orig_helm_visitor, helm_visitor)
        diff_resource_keys = get_helm_release_diff_keys(
            orig_content, content, dependency_map
        )
        diff_names = {
            resource_key.namespaced_name for resource_key in diff_resource_keys
        }
        helm_visitor.releases = [
            release
            for release in helm_visitor.releases
            if release.namespaced_name in diff_names
        ]
        orig_helm_visitor.releases = [
            release
            for release in orig_helm_visitor.releases
            if release.namespaced_name in diff_names
        ]

        helm_content = ObjectOutput(strip_attrs)
        orig_helm_content = ObjectOutput(strip_attrs)
        with tempfile.TemporaryDirectory() as helm_cache_dir:
            await asyncio.gather(
                helm_visitor.inflate(
                    pathlib.Path(helm_cache_dir), helm_content.visitor(), options
                ),
                orig_helm_visitor.inflate(
                    pathlib.Path(helm_cache_dir), orig_helm_content.visitor(), options
                ),
            )

        with open(output_file, "w") as file:
            if output == "yaml":
                for line in perform_yaml_diff(
                    orig_helm_content, helm_content, unified, limit_bytes
                ):
                    print(line, file=file)
            elif output == "json":
                for line in perform_json_diff(
                    orig_helm_content, helm_content, unified, limit_bytes
                ):
                    print(line, file=file)
            elif external_diff := os.environ.get("DIFF"):
                async for line in perform_external_diff(
                    shlex.split(external_diff),
                    orig_helm_content,
                    helm_content,
                    limit_bytes,
                ):
                    print(line, file=file)
            else:
                for line in perform_object_diff(
                    orig_helm_content, helm_content, unified, limit_bytes
                ):
                    print(line, file=file)


class DiffAction:
    """Flux-local diff action."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args: ArgumentParser = subparsers.add_parser(
            "diff",
            help="Diff a local flux resource",
            description="""You may also use flux-local to verify your local changes
                to cluster resources have the desird effect. This is similar to flux
                diff but entirely local. This will run a local kustomize build first
                against the local repo then again against a prior repo revision, then
                prints the output.""",
        )
        subcmds = args.add_subparsers(
            title="Available commands",
            required=True,
        )
        DiffKustomizationAction.register(subcmds)
        DiffHelmReleaseAction.register(subcmds)
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        # No-op given subcommands are dispatched



"""Library for issuing commands using asyncio and returning the result."""

import asyncio
from abc import ABC, abstractmethod
import logging
import shlex
import subprocess
from dataclasses import dataclass
from pathlib import Path
from collections.abc import Sequence
import os

from .exceptions import CommandException

_LOGGER = logging.getLogger(__name__)

_CONCURRENCY = 20
_SEM = asyncio.Semaphore(_CONCURRENCY)
_TIMEOUT = 20.0


# No public API
__all__: list[str] = []


class Task(ABC):
    """An instance of a async task to execute."""

    @abstractmethod
    async def run(self, stdin: bytes | None = None) -> bytes:
        """Execute the task and return the result."""


def format_path(path: Path) -> str:
    """Format path for debugging."""
    if path.is_absolute():
        cwd = Path.cwd()
        if path.is_relative_to(cwd):
            rel_path = str(path.relative_to(cwd))
            return f"{rel_path} (abs)"
    return str(path)


@dataclass
class Command(Task):
    """An instance of a command to run."""

    cmd: list[str]
    """Array of command line arguments."""

    cwd: Path | None = None
    """Current working directory."""

    exc: type[CommandException] = CommandException
    """Exception to throw in case of an error."""

    retcodes: list[int] | None = None
    """Non-zero error codes that are allowed to indicate success (e.g. for diff)."""

    env: dict[str, str] | None = None
    """Environment variables for the subprocess."""

    @property
    def string(self) -> str:
        """Render the command as a single string."""
        return " ".join([shlex.quote(arg) for arg in self.cmd])

    def __str__(self) -> str:
        """Render as a debug string."""
        cwd: str = ""
        if self.cwd:
            cwd = f"({format_path(self.cwd)}) "
        return f"{cwd}{self.string}"

    async def run(self, stdin: bytes | None = None) -> bytes:
        """Run the command, returning stdout."""
        _LOGGER.debug("Running command: %s", self)
        env = {
            **os.environ,
            **(self.env if self.env else {}),
        }
        proc = await asyncio.create_subprocess_shell(
            self.string,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.cwd,
            env=env,
        )
        out, err = await proc.communicate(stdin)
        if proc.returncode:
            if self.retcodes and proc.returncode in self.retcodes:
                return out
            errors = [f"Command '{self}' failed with return code {proc.returncode}"]
            if out:
                errors.append(out.decode("utf-8"))
            if err:
                errors.append(err.decode("utf-8"))
            _LOGGER.debug("\n".join(errors))
            raise self.exc("\n".join(errors))
        return out


async def _run_piped_with_sem(cmds: Sequence[Task]) -> str:
    """Run a set of commands, piped together, returning stdout of last."""
    stdin = None
    out = None
    for cmd in cmds:
        try:
            out = await asyncio.wait_for(cmd.run(stdin), _TIMEOUT)
        except asyncio.exceptions.TimeoutError as err:
            if isinstance(cmd, Command):
                raise cmd.exc(f"Command '{cmd}' timed out") from err
            raise err
        stdin = out
    return out.decode("utf-8") if out else ""


async def run_piped(cmds: Sequence[Task]) -> str:
    """Run a set of commands, piped together, returning stdout of last."""
    async with _SEM:
        result = await _run_piped_with_sem(cmds)
    return result


async def run(cmd: Task) -> str:
    """Run the specified command and return stdout."""
    return await run_piped([cmd])



"""Tests for the resource diff module."""

import pathlib
import shutil
import tempfile
from contextlib import contextmanager
from collections.abc import Generator

import yaml
from syrupy.assertion import SnapshotAssertion

from flux_local import git_repo
from flux_local.resource_diff import (
    get_helm_release_diff_keys,
    merge_helm_releases,
    merge_named_resources,
    build_helm_dependency_map,
    perform_json_diff,
    perform_yaml_diff,
)
from flux_local.visitor import ObjectOutput, HelmVisitor, ResourceKey
from flux_local.manifest import HelmRelease, HelmChart, NamedResource

TESTDATA_PATH = pathlib.Path("tests/testdata/cluster8")


async def visit_helm_content(path: pathlib.Path) -> tuple[ObjectOutput, HelmVisitor]:
    """Visit content in the specified path."""

    query = git_repo.ResourceSelector()
    query.path = git_repo.PathSelector(path)
    query.helm_release.namespace = None  # All

    content = ObjectOutput([])
    query.kustomization.visitor = content.visitor()

    helm_visitor = HelmVisitor()
    query.helm_repo.visitor = helm_visitor.repo_visitor()
    query.helm_release.visitor = helm_visitor.release_visitor()

    await git_repo.build_manifest(selector=query)

    return content, helm_visitor


@contextmanager
def mirror_worktree(path: pathlib.Path) -> Generator[pathlib.Path]:
    """Create a copy of the path in a temporary git repo for generating diffs."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = pathlib.Path(tmpdir)
        # Required by `git_repo` to find the root
        shutil.copytree(".git", str(tmp_path / ".git"))

        new_path = tmp_path / path
        shutil.copytree(path, new_path)

        yield new_path


def test_no_diff_keys_empty_input() -> None:
    """Test get_helm_release_diff_keys with empty inputs."""

    a = ObjectOutput([])
    b = ObjectOutput([])
    assert get_helm_release_diff_keys(a, b, {}) == []


async def test_no_diff_keys() -> None:
    """Test get_helm_release_diff_keys with no diff."""

    content, helm_visitor = await visit_helm_content(TESTDATA_PATH)
    with mirror_worktree(TESTDATA_PATH) as new_cluster_path:
        content_new, helm_visitor_new = await visit_helm_content(new_cluster_path)

    dep_map = build_helm_dependency_map(helm_visitor, helm_visitor_new)
    assert get_helm_release_diff_keys(content, content_new, dep_map) == []


async def test_helm_release_spec_diff() -> None:
    """Test get_helm_release_diff_keys with a diff in the input values."""

    content, helm_visitor = await visit_helm_content(TESTDATA_PATH)
    with mirror_worktree(TESTDATA_PATH) as new_cluster_path:

        # Generate a diff by changing a value in the HelmRelease
        podinfo_path = new_cluster_path / "apps/podinfo.yaml"
        podinfo_doc = list(
            yaml.load_all(podinfo_path.read_text(), Loader=yaml.SafeLoader)
        )
        assert len(podinfo_doc) == 2
        assert podinfo_doc[1]["kind"] == "HelmRelease"
        assert podinfo_doc[1]["spec"]["values"]["redis"]["enabled"]
        podinfo_doc[1]["spec"]["values"]["redis"]["enabled"] = False
        podinfo_path.write_text(yaml.dump_all(podinfo_doc))

        content_new, helm_visitor_new = await visit_helm_content(new_cluster_path)

    dep_map = build_helm_dependency_map(helm_visitor, helm_visitor_new)
    assert get_helm_release_diff_keys(content, content_new, dep_map) == [
        ResourceKey(
            kustomization_path=str(TESTDATA_PATH / "apps"),
            kind="HelmRelease",
            namespace="podinfo",
            name="podinfo",
        )
    ]


async def test_helm_repository_causes_diff() -> None:
    """Test get_helm_release_diff_keys where an input HelmRepository change causes a diff."""

    content, helm_visitor = await visit_helm_content(TESTDATA_PATH)
    with mirror_worktree(TESTDATA_PATH) as new_cluster_path:

        # Generate a diff by changing a value in the HelmRelease
        podinfo_path = new_cluster_path / "apps/podinfo.yaml"
        podinfo_doc = list(
            yaml.load_all(podinfo_path.read_text(), Loader=yaml.SafeLoader)
        )
        assert len(podinfo_doc) == 2
        assert podinfo_doc[0]["kind"] == "HelmRepository"
        assert podinfo_doc[0]["spec"]["url"] == "oci://ghcr.io/stefanprodan/charts"
        podinfo_doc[0]["spec"]["url"] = "oci://ghcr.io/stefanprodan/charts2"
        podinfo_path.write_text(yaml.dump_all(podinfo_doc))

        content_new, helm_visitor_new = await visit_helm_content(new_cluster_path)

    dep_map = build_helm_dependency_map(helm_visitor, helm_visitor_new)
    assert get_helm_release_diff_keys(content, content_new, dep_map) == [
        ResourceKey(
            kustomization_path=str(TESTDATA_PATH / "apps"),
            kind="HelmRelease",
            namespace="podinfo",
            name="podinfo",
        )
    ]


async def test_merge_helm_releases() -> None:
    """Test merge_helm_releases."""

    chart = HelmChart(
        name="chart", version="1.0.0", repo_name="example", repo_namespace="flux-system"
    )

    a = [
        HelmRelease(
            name="a", namespace="default", chart=chart, values={"key": "value1"}
        ),
        HelmRelease(
            name="b", namespace="default", chart=chart, values={"key": "value2"}
        ),
    ]
    b = [
        HelmRelease(
            name="b", namespace="default", chart=chart, values={"key": "value3"}
        ),
        HelmRelease(
            name="c", namespace="default", chart=chart, values={"key": "value4"}
        ),
    ]
    assert list(merge_helm_releases(a, b)) == [
        (
            HelmRelease(
                name="a", namespace="default", chart=chart, values={"key": "value1"}
            ),
            None,
        ),
        (
            HelmRelease(
                name="b", namespace="default", chart=chart, values={"key": "value2"}
            ),
            HelmRelease(
                name="b", namespace="default", chart=chart, values={"key": "value3"}
            ),
        ),
        (
            None,
            HelmRelease(
                name="c", namespace="default", chart=chart, values={"key": "value4"}
            ),
        ),
    ]


async def test_merge_named_resources() -> None:
    """Test for merging named resources."""
    a = [
        NamedResource(name="podinfo", namespace="podinfo", kind="HelmRelease"),
        NamedResource(name="podinfo", namespace="podinfo", kind="ConfigMap"),
        NamedResource(name="secret", namespace="podinfo", kind="Secret"),
    ]
    b = [
        NamedResource(name="podinfo", namespace="podinfo", kind="HelmRelease"),
        NamedResource(name="podinfo", namespace="podinfo", kind="ConfigMap"),
        NamedResource(name="podinfo", namespace="flux-system", kind="HelmRepository"),
    ]
    assert list(merge_named_resources(a, b)) == [
        NamedResource(name="podinfo", namespace="podinfo", kind="HelmRelease"),
        NamedResource(name="podinfo", namespace="podinfo", kind="ConfigMap"),
        NamedResource(name="secret", namespace="podinfo", kind="Secret"),
        NamedResource(name="podinfo", namespace="flux-system", kind="HelmRepository"),
    ]


async def test_perform_yaml_diff(snapshot: SnapshotAssertion) -> None:
    """Test perform_yaml_diff where an input HelmRepository change causes a diff."""

    content, helm_visitor = await visit_helm_content(TESTDATA_PATH)
    with mirror_worktree(TESTDATA_PATH) as new_cluster_path:

        # Generate a diff by changing a value in the HelmRelease
        podinfo_path = new_cluster_path / "apps/podinfo.yaml"
        podinfo_doc = list(
            yaml.load_all(podinfo_path.read_text(), Loader=yaml.SafeLoader)
        )
        assert len(podinfo_doc) == 2
        assert podinfo_doc[0]["kind"] == "HelmRepository"
        assert podinfo_doc[0]["spec"]["url"] == "oci://ghcr.io/stefanprodan/charts"
        podinfo_doc[0]["spec"]["url"] = "oci://ghcr.io/stefanprodan/charts2"
        podinfo_path.write_text(yaml.dump_all(podinfo_doc))

        content_new, helm_visitor_new = await visit_helm_content(new_cluster_path)

    diff = "\n".join(list(perform_yaml_diff(content, content_new, 5, 50000)))
    assert diff == snapshot


async def test_perform_json_diff(snapshot: SnapshotAssertion) -> None:
    """Test perform_yaml_diff where an input HelmRepository change causes a diff."""

    content, helm_visitor = await visit_helm_content(TESTDATA_PATH)
    with mirror_worktree(TESTDATA_PATH) as new_cluster_path:

        # Generate a diff by changing a value in the HelmRelease
        podinfo_path = new_cluster_path / "apps/podinfo.yaml"
        podinfo_doc = list(
            yaml.load_all(podinfo_path.read_text(), Loader=yaml.SafeLoader)
        )
        assert len(podinfo_doc) == 2
        assert podinfo_doc[0]["kind"] == "HelmRepository"
        assert podinfo_doc[0]["spec"]["url"] == "oci://ghcr.io/stefanprodan/charts"
        podinfo_doc[0]["spec"]["url"] = "oci://ghcr.io/stefanprodan/charts2"
        podinfo_path.write_text(yaml.dump_all(podinfo_doc))

        content_new, helm_visitor_new = await visit_helm_content(new_cluster_path)

    diff = "\n".join(list(perform_json_diff(content, content_new, 5, 50000)))
    assert diff == snapshot



"""Module for computing resource diffs.

This is used internally, primarily by the diff tool.
"""

from collections.abc import Iterable, Callable
from dataclasses import asdict
import difflib
import logging
import pathlib
import tempfile
from typing import Generator, Any, AsyncGenerator, TypeVar
import yaml
import json


from . import command
from .visitor import ObjectOutput, ResourceKey, HelmVisitor
from .manifest import HelmRelease, NamedResource

_LOGGER = logging.getLogger(__name__)

_TRUNCATE = "[Diff truncated by flux-local]"

T = TypeVar("T")


def _unique_keys(k1: dict[T, Any], k2: dict[T, Any]) -> Iterable[T]:
    """Return an ordered set."""
    return {
        **{k: True for k in k1.keys()},
        **{k: True for k in k2.keys()},
    }.keys()


def perform_object_diff(
    a: ObjectOutput, b: ObjectOutput, n: int, limit_bytes: int
) -> Generator[str, None, None]:
    """Generate diffs between the two output objects."""
    for kustomization_key in _unique_keys(a.content, b.content):
        _LOGGER.debug(
            "Diffing results for Kustomization %s (n=%d)", kustomization_key, n
        )
        a_resources = a.content.get(kustomization_key, {})
        b_resources = b.content.get(kustomization_key, {})
        for resource_key in _unique_keys(a_resources, b_resources):
            diff_text = difflib.unified_diff(
                a=a_resources.get(resource_key, []),
                b=b_resources.get(resource_key, []),
                fromfile=f"{kustomization_key.label} {resource_key.compact_label}",
                tofile=f"{kustomization_key.label} {resource_key.compact_label}",
                n=n,
            )
            size = 0
            for line in diff_text:
                size += len(line)
                if limit_bytes and size > limit_bytes:
                    yield _TRUNCATE
                    break
                yield line


async def perform_external_diff(
    cmd: list[str],
    a: ObjectOutput,
    b: ObjectOutput,
    limit_bytes: int,
) -> AsyncGenerator[str, None]:
    """Generate diffs between the two output objects."""
    with tempfile.TemporaryDirectory() as tmpdir:
        for kustomization_key in _unique_keys(a.content, b.content):
            _LOGGER.debug(
                "Diffing results for Kustomization %s",
                kustomization_key,
            )
            a_resources = a.content.get(kustomization_key, {})
            b_resources = b.content.get(kustomization_key, {})
            keys = _unique_keys(a_resources, b_resources)

            a_file = pathlib.Path(tmpdir) / "a.yaml"
            a_file.write_text(
                "\n".join(
                    [
                        "\n".join(a_resources.get(resource_key, []))
                        for resource_key in keys
                    ]
                )
            )
            b_file = pathlib.Path(tmpdir) / "b.yaml"
            b_file.write_text(
                "\n".join(
                    [
                        "\n".join(b_resources.get(resource_key, []))
                        for resource_key in keys
                    ]
                )
            )

            out = await command.Command(
                cmd + [str(a_file), str(b_file)], retcodes=[0, 1]
            ).run()
            if out:
                result = out.decode("utf-8")
                if limit_bytes and len(result) > limit_bytes:
                    result = result[:limit_bytes] + "\n" + _TRUNCATE
                yield result


def _omit_none(obj: Any) -> dict[str, Any]:
    """Creates a dictionary with None values missing."""
    return {k: v for k, v in obj if v is not None}


def perform_yaml_diff(
    a: ObjectOutput,
    b: ObjectOutput,
    n: int,
    limit_bytes: int,
) -> Generator[str, None, None]:
    """Generate diffs between the two output objects."""

    def diff_func(diffs: list[dict[str, Any]]) -> str:
        return yaml.dump(
            diffs, sort_keys=False, explicit_start=True, default_style=None
        )

    for result in _perform_function_diff(a, b, n, limit_bytes, diff_func):
        yield result


def perform_json_diff(
    a: ObjectOutput,
    b: ObjectOutput,
    n: int,
    limit_bytes: int,
) -> Generator[str, None, None]:
    """Generate diffs between the two output objects."""

    def diff_func(diffs: list[dict[str, Any]]) -> str:
        return json.dumps(diffs, sort_keys=False, indent=4)

    for result in _perform_function_diff(a, b, n, limit_bytes, diff_func):
        yield result


def _perform_function_diff(
    a: ObjectOutput,
    b: ObjectOutput,
    n: int,
    limit_bytes: int,
    diff_func: Callable[[list[dict[str, Any]]], str],
) -> Generator[str, None, None]:
    """Generate diffs between the two output objects."""

    diffs: list[dict[str, Any]] = []
    for kustomization_key in _unique_keys(a.content, b.content):
        _LOGGER.debug("Diffing results for %s (n=%d)", kustomization_key, n)
        a_resources = a.content.get(kustomization_key, {})
        b_resources = b.content.get(kustomization_key, {})
        resource_diffs = []
        for resource_key in _unique_keys(a_resources, b_resources):
            diff_text = difflib.unified_diff(
                a=a_resources.get(resource_key, []),
                b=b_resources.get(resource_key, []),
                fromfile=f"{kustomization_key.label} {resource_key.compact_label}",
                tofile=f"{kustomization_key.label} {resource_key.compact_label}",
                n=n,
            )
            diff_content = "\n".join(diff_text)
            if not diff_content:
                continue
            if limit_bytes and len(diff_content) > limit_bytes:
                diff_content = diff_content[:limit_bytes] + "\n" + _TRUNCATE
            obj = {
                **asdict(resource_key, dict_factory=_omit_none),
                "diff": diff_content,
            }
            resource_diffs.append(obj)
        if resource_diffs:
            diffs.append(
                {
                    **asdict(kustomization_key),
                    "diffs": resource_diffs,
                }
            )
    if diffs:
        yield diff_func(diffs)


def merge_helm_releases(
    a: list[HelmRelease], b: list[HelmRelease]
) -> Generator[tuple[HelmRelease | None, HelmRelease | None], None, None]:
    """Return HelmReleases joined by name."""
    a_dict = {r.release_name: r for r in a}
    b_dict = {r.release_name: r for r in b}
    for name in _unique_keys(a_dict, b_dict):
        yield a_dict.get(name), b_dict.get(name)


def merge_named_resources(
    a: list[NamedResource], b: list[NamedResource]
) -> Generator[NamedResource, None, None]:
    """Return HelmReleases joined by name."""
    a_dict = {f"{r.kind}-{r.namespaced_name}": r for r in a}
    b_dict = {f"{r.kind}-{r.namespaced_name}": r for r in b}
    for name in _unique_keys(a_dict, b_dict):
        # Emit either the left or right since they should be identical
        value: NamedResource = a_dict.get(name) or b_dict.get(name)  # type: ignore[assignment]
        yield value


def build_helm_dependency_map(
    a_visitor: HelmVisitor, b_visitor: HelmVisitor
) -> dict[NamedResource, NamedResource]:
    """Return a map of all resources to the HelmRelease resource that depends on them.

    This is a mapping of all the named resources. If any of them changed, then we
    need to diff the HelmRelease to look for changes in the rendered output.
    """
    results: dict[NamedResource, NamedResource] = {}
    for helmrelease_a, helmrelease_b in merge_helm_releases(
        a_visitor.releases, b_visitor.releases
    ):
        resources_a = helmrelease_a.resource_dependencies if helmrelease_a else []
        resources_b = helmrelease_b.resource_dependencies if helmrelease_b else []
        hr: HelmRelease = helmrelease_a or helmrelease_b  # type: ignore[assignment]
        hr_named_resource = NamedResource(
            kind="HelmRelease",
            namespace=hr.namespace,
            name=hr.name,
        )
        for named_resource in merge_named_resources(resources_a, resources_b):
            results[named_resource] = hr_named_resource

    return results


def get_helm_release_diff_keys(
    a: ObjectOutput, b: ObjectOutput, dependency_map: dict[NamedResource, NamedResource]
) -> list[ResourceKey]:
    """Return HelmRelease resource keys with diffs, by cluster."""

    # Pass #1: Check all named resources that are upstream dependencies of a
    # HelmRelease and have a content change. Save the HelmRelease name.
    hrs_to_check: set[NamedResource] = set()
    _LOGGER.debug("Checking HelmRelease dependencies")
    for kustomization_key in _unique_keys(a.content, b.content):
        _LOGGER.debug("Diffing results for Kustomization %s", kustomization_key)
        a_resources = a.content.get(kustomization_key, {})
        b_resources = b.content.get(kustomization_key, {})
        for resource_key in _unique_keys(a_resources, b_resources):
            if (
                hr_named_resource := dependency_map.get(resource_key.named_resource)
            ) is None:
                continue
            if a_resources.get(resource_key) != b_resources.get(resource_key):
                _LOGGER.info(
                    "HelmRelease %s detected potential change in resource %s (Kustomization %s)",
                    hr_named_resource.namespaced_name,
                    resource_key.namespaced_name,
                    kustomization_key.namespaced_name,
                )
                hrs_to_check.add(hr_named_resource)

    # Pass #2: Find the ResourceKey of all the changed HelmReleases
    results: list[ResourceKey] = []
    for kustomization_key in _unique_keys(a.content, b.content):
        a_resources = a.content.get(kustomization_key, {})
        b_resources = b.content.get(kustomization_key, {})
        for resource_key in _unique_keys(a_resources, b_resources):
            if resource_key.named_resource in hrs_to_check:
                results.append(resource_key)

    return results



"""Helper functions for working with container images."""

import logging
from typing import Any

from . import git_repo, manifest

_LOGGER = logging.getLogger(__name__)


# Object types that may have container images.
KINDS = [
    "Pod",
    "Deployment",
    "StatefulSet",
    "ReplicaSet",
    "DaemonSet",
    "CronJob",
    "Job",
    "ReplicationController",
    "EMQX",                  # apps.emqx.io/v2beta1
    "Cluster",               # postgresql.cnpg.io/v1
    "CephCluster",           # ceph.rook.io/v1
    "Alertmanager",          # monitoring.coreos.com/v1
    "Prometheus",            # monitoring.coreos.com/v1
    "AutoscalingRunnerSet",  # actions.github.com/v1alpha1
]

# Default image key for most object types.
IMAGE_KEY = "image"

# Override the default image key for some object types.
KINDS_IMAGE_KEY = {
    "Cluster": "imageName"
}


def _extract_images(kind: str, doc: dict[str, Any]) -> set[str]:
    """Extract the image from a Kubernetes object."""
    images: set[str] = set({})
    image_key = KINDS_IMAGE_KEY.get(kind) or IMAGE_KEY

    for key, value in doc.items():
        if key == image_key:
            images.add(value)
        elif isinstance(value, dict):
            images.update(_extract_images(kind, value))
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    images.update(_extract_images(kind, item))

    return images


class ImageVisitor:
    """Helper that visits container image related objects.

    This tracks the container images used by the kustomizations and HelmReleases
    so they can be dumped for further verification.
    """

    def __init__(self) -> None:
        """Initialize ImageVisitor."""
        self.images: dict[str, set[str]] = {}

    def repo_visitor(self) -> git_repo.DocumentVisitor:
        """Return a git_repo.DocumentVisitor that points to this object."""

        def add_image(name: str, doc: dict[str, Any]) -> None:
            """Visitor function to find relevant images and record them for later inspection.

            Updates the image set with the images found in the document.
            """
            kind: str = doc["kind"]
            images = _extract_images(kind, doc)
            if not images:
                return
            if name in self.images:
                self.images[name].update(images)
            else:
                self.images[name] = set(images)

        return git_repo.DocumentVisitor(kinds=KINDS, func=add_image)

    def update_manifest(self, manifest: manifest.Manifest) -> None:
        """Update the manifest with the images found in the repo."""
        for cluster in manifest.clusters:
            for kustomization in cluster.kustomizations:
                if images := self.images.get(kustomization.namespaced_name):
                    kustomization.images = list(images)
                    kustomization.images.sort()



"""Tests for the visitor module."""

from typing import Any
import yaml

import pytest


from flux_local.visitor import strip_resource_attributes

STRIP_ATTRIBUTES = [
    "app.kubernetes.io/version",
    "chart",
]


@pytest.mark.parametrize(
    ("metadata", "expected_metadata"),
    [
        (
            {
                "labels": {
                    "app.kubernetes.io/version": "1.0.0",
                    "app.kubernetes.io/managed-by": "Helm",
                }
            },
            {
                "labels": {
                    "app.kubernetes.io/managed-by": "Helm",
                },
            },
        ),
        (
            {
                "annotations": {
                    "app.kubernetes.io/version": "1.0.0",
                    "app.kubernetes.io/managed-by": "Helm",
                }
            },
            {
                "annotations": {
                    "app.kubernetes.io/managed-by": "Helm",
                },
            },
        ),
        (
            {},
            {},
        ),
    ],
)
def test_strip_resource_attributes(
    metadata: dict[str, Any], expected_metadata: dict[str, Any]
) -> None:
    """Test the strip_resource_attributes function."""
    resource = {
        "apiVersion": "v1",
        "kind": "ConfigMap",
        "metadata": {
            "name": "my-configmap",
            "namespace": "default",
            **metadata,
        },
        "data": {
            "key1": "value1",
            "key2": "value2",
        },
    }
    strip_resource_attributes(resource, STRIP_ATTRIBUTES)
    assert resource == {
        "apiVersion": "v1",
        "kind": "ConfigMap",
        "metadata": {
            "name": "my-configmap",
            "namespace": "default",
            **expected_metadata,
        },
        "data": {
            "key1": "value1",
            "key2": "value2",
        },
    }


def test_strip_deployment_metadata() -> None:
    """Test the strip_resource_attributes function."""
    resource = yaml.load(
        """apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-deployment
  labels:
    app: nginx
spec:
  replicas: 3
  selector:
    matchLabels:
      app: nginx
  template:
    metadata:
      labels:
        app: nginx
        app.kubernetes.io/version: 1.0.0
    spec:
      containers:
      - name: nginx
        image: nginx:1.14.2
        ports:
        - containerPort: 80
""",
        Loader=yaml.Loader,
    )

    strip_resource_attributes(resource, STRIP_ATTRIBUTES)
    assert (
        yaml.dump(resource, sort_keys=False)
        == """apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-deployment
  labels:
    app: nginx
spec:
  replicas: 3
  selector:
    matchLabels:
      app: nginx
  template:
    metadata:
      labels:
        app: nginx
    spec:
      containers:
      - name: nginx
        image: nginx:1.14.2
        ports:
        - containerPort: 80
"""
    )


def test_strip_list_metadata() -> None:
    """Test the stripping metadata from a list resource."""
    resource = yaml.load(
        """apiVersion: v1
items:
- apiVersion: stable.example.com/v1
  kind: CronTab
  metadata:
    annotations:
      app: my-cron-tab
      app.kubernetes.io/version: 1.0.0
    creationTimestamp: '2021-06-20T07:35:27Z'
    generation: 1
    name: my-new-cron-object
    namespace: default
    resourceVersion: '1326'
    uid: 9aab1d66-628e-41bb-a422-57b8b3b1f5a9
  spec:
    cronSpec: '* * * * */5'
    image: my-awesome-cron-image
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''

""",
        Loader=yaml.Loader,
    )

    strip_resource_attributes(resource, STRIP_ATTRIBUTES)
    assert (
        yaml.dump(resource, sort_keys=False)
        == """apiVersion: v1
items:
- apiVersion: stable.example.com/v1
  kind: CronTab
  metadata:
    annotations:
      app: my-cron-tab
    creationTimestamp: '2021-06-20T07:35:27Z'
    generation: 1
    name: my-new-cron-object
    namespace: default
    resourceVersion: '1326'
    uid: 9aab1d66-628e-41bb-a422-57b8b3b1f5a9
  spec:
    cronSpec: '* * * * */5'
    image: my-awesome-cron-image
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''
"""
    )


def test_strip_list_null_items() -> None:
    """Test corner cases of handling metadata."""
    resource = yaml.load(
        """apiVersion: v1
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''
items:

""",
        Loader=yaml.Loader,
    )

    strip_resource_attributes(resource, STRIP_ATTRIBUTES)
    assert (
        yaml.dump(resource, sort_keys=False)
        == """apiVersion: v1
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''
items: null
"""
    )


def test_strip_list_item_without_metdata() -> None:
    """Test corner cases of handling metadata."""
    resource = yaml.load(
        """apiVersion: v1
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''
items:
- kind: CronTab
""",
        Loader=yaml.Loader,
    )

    strip_resource_attributes(resource, STRIP_ATTRIBUTES)
    assert (
        yaml.dump(resource, sort_keys=False)
        == """apiVersion: v1
kind: List
metadata:
  resourceVersion: ''
  selfLink: ''
items:
- kind: CronTab
"""
    )



"""Library for running `helm template` to produce local items in the cluster.

You can instantiate a helm template with the following:
- A HelmRepository which is a url that contains charts
- A HelmRelease which is an instance of a HelmChart in a HelmRepository

This is an example that prepares the helm repository:
```python
from flux_local.kustomize import Kustomize
from flux_local.helm import Helm
from flux_local.manifest import HelmRepository

kustomize = Kustomize.build(TESTDATA_DIR)
repos = await kustomize.grep("kind=^HelmRepository$").objects()
helm = Helm("/tmp/path/helm", "/tmp/path/cache")
for repo in repos:
  helm.add_repo(HelmRepository.parse_doc(repo))
await helm.update()
```

Then to actually instantiate a template from a HelmRelease:
```python
from flux_local.manifest import HelmRelease

releases = await kustomize.grep("kind=^HelmRelease$").objects()
if not len(releases) == 1:
    raise ValueError("Expected only one HelmRelease")
tmpl = helm.template(HelmRelease.parse_doc(releases[0]))
objects = await tmpl.objects()
for object in objects:
    print(f"Found object {object['apiVersion']} {object['kind']}")
```
"""

import datetime
from dataclasses import dataclass
import logging
from pathlib import Path
from typing import Any

import aiofiles
import yaml

from . import command
from .kustomize import Kustomize
from .manifest import (
    HelmRelease,
    HelmRepository,
    OCIRepository,
    CRD_KIND,
    SECRET_KIND,
    REPO_TYPE_OCI,
    HELM_REPOSITORY,
    GIT_REPOSITORY,
    OCI_REPOSITORY,
)
from .exceptions import HelmException

__all__ = [
    "Helm",
    "Options",
]

_LOGGER = logging.getLogger(__name__)


HELM_BIN = "helm"
DEFAULT_REGISTRY_CONFIG = "/dev/null"


def _chart_name(
    release: HelmRelease, repo: HelmRepository | OCIRepository | None
) -> str:
    """Return the helm chart name used for the helm template command."""
    if release.chart.repo_kind == OCI_REPOSITORY:
        assert repo
        if isinstance(repo, OCIRepository):
            return repo.url
        raise HelmException(
            f"HelmRelease {release.name} expected OCIRepository but got HelmRepository {repo.repo_name}"
        )
    if release.chart.repo_kind == HELM_REPOSITORY:
        assert repo
        if not isinstance(repo, HelmRepository):
            raise HelmException(
                f"HelmRelease {release.name} expected HelmRepository but got OCIRepository {repo.repo_name}"
            )
        if repo and repo.repo_type == REPO_TYPE_OCI:
            return f"{repo.url}/{release.chart.name}"
        return release.chart.chart_name
    elif release.chart.repo_kind == GIT_REPOSITORY:
        return release.chart.name
    raise HelmException(
        f"Unable to find chart source for chart {release.chart.chart_name} "
        f"kind {release.chart.repo_kind} for HelmRelease {release.name}"
    )


class RepositoryConfig:
    """Generates a helm repository configuration from flux HelmRepository objects."""

    def __init__(self, repos: list[HelmRepository]) -> None:
        """Initialize RepositoryConfig."""
        self._repos = repos

    @property
    def config(self) -> dict[str, Any]:
        """Return a synthetic repository config object."""
        now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        return {
            "apiVersion": "",
            "generated": now.isoformat(),
            "repositories": [
                {
                    "name": f"{repo.namespace}-{repo.name}",
                    "url": repo.url,
                }
                for repo in self._repos
            ],
        }


@dataclass
class Options:
    """Options to use when inflating a Helm chart.

    Internally, these translate into either command line flags or
    resource kinds that will be filtered form the output.
    """

    skip_crds: bool = True
    """Skip CRDs when building the output."""

    skip_tests: bool = True
    """Don't run helm tests on the output."""

    skip_secrets: bool = False
    """Don't emit secrets in the output."""

    skip_kinds: list[str] | None = None
    """Omit these kinds in the output."""

    kube_version: str | None = None
    """Value of the helm --kube-version flag."""

    api_versions: str | None = None
    """Value of the helm --api-versions flag."""

    registry_config: str | None = None
    """Value of the helm --registry-config flag."""

    @property
    def base_args(self) -> list[str]:
        """Helm template CLI arguments built from the options."""
        return ["--registry-config", self.registry_config or DEFAULT_REGISTRY_CONFIG]

    @property
    def template_args(self) -> list[str]:
        """Helm template CLI arguments built from the options."""
        args = self.base_args
        if self.skip_crds:
            args.append("--skip-crds")
        if self.skip_tests:
            args.append("--skip-tests")
        if self.kube_version:
            args.extend(["--kube-version", self.kube_version])
        if self.api_versions:
            args.extend(["--api-versions", self.api_versions])
        return args

    @property
    def skip_resources(self) -> list[str]:
        """A list of CRD resources to filter from the output."""
        skips = []
        if self.skip_crds:
            skips.append(CRD_KIND)
        if self.skip_secrets:
            skips.append(SECRET_KIND)
        if self.skip_kinds:
            skips.extend(self.skip_kinds)
        return skips


class Helm:
    """Manages local HelmRepository state."""

    def __init__(self, tmp_dir: Path, cache_dir: Path) -> None:
        """Initialize Helm."""
        self._tmp_dir = tmp_dir
        self._repo_config_file = self._tmp_dir / "repository-config.yaml"
        self._flags = [
            "--repository-cache",
            str(cache_dir),
            "--repository-config",
            str(self._repo_config_file),
        ]
        self._repos: list[HelmRepository | OCIRepository] = []

    def add_repo(self, repo: HelmRepository | OCIRepository) -> None:
        """Add the specified HelmRepository to the local config."""
        self._repos.append(repo)

    def add_repos(
        self,
        repos: (
            list[HelmRepository]
            | list[OCIRepository]
            | list[HelmRepository | OCIRepository]
        ),
    ) -> None:
        """Add the specified HelmRepository to the local config."""
        for repo in repos:
            self._repos.append(repo)

    async def update(self) -> None:
        """Run the update command to update the local repo.

        Typically the repository must be updated before doing any chart templating.
        """
        _LOGGER.debug("Updating %d repositories", len(self._repos))
        repos = [
            repo
            for repo in self._repos
            if isinstance(repo, HelmRepository) and repo.repo_type != REPO_TYPE_OCI
        ]
        if not repos:
            return
        content = yaml.dump(RepositoryConfig(repos).config, sort_keys=False)
        async with aiofiles.open(str(self._repo_config_file), mode="w") as config_file:
            await config_file.write(content)
        args = [HELM_BIN, "repo", "update"]
        args.extend(Options().base_args)
        args.extend(self._flags)
        await command.run(command.Command(args, exc=HelmException))

    async def template(
        self,
        release: HelmRelease,
        options: Options | None = None,
    ) -> Kustomize:
        """Return command to evaluate the template of the specified chart.

        The values will come from the `HelmRelease` object.
        """
        if options is None:
            options = Options()
        repo = next(
            iter([repo for repo in self._repos if repo.repo_name == release.repo_name]),
            None,
        )
        # We'll attempt to make a chart name for a GitRepository below and it will
        # be somewhat best effort.
        if not repo and release.chart.repo_kind != GIT_REPOSITORY:
            raise HelmException(
                f"Unable to find HelmRepository or OCIRepository for {release.chart.chart_name} for "
                f"HelmRelease {release.name} "
                f"({len(self._repos)} other HelmRepositories/OCIRepositories in --path)"
            )
        args: list[str] = [
            HELM_BIN,
            "template",
            release.name,
            _chart_name(release, repo),
            "--namespace",
            release.namespace,
        ]
        args.extend(self._flags)
        args.extend(options.template_args)
        if release.chart.version:
            args.extend(
                [
                    "--version",
                    release.chart.version,
                ]
            )
        elif isinstance(repo, OCIRepository) and repo.ref_tag:
            args.extend(
                [
                    "--version",
                    repo.ref_tag,
                ]
            )
        if release.values:
            values_path = self._tmp_dir / f"{release.release_name}-values.yaml"
            async with aiofiles.open(values_path, mode="w") as values_file:
                await values_file.write(yaml.dump(release.values, sort_keys=False))
            args.extend(["--values", str(values_path)])
        cmd = Kustomize([command.Command(args, exc=HelmException)])
        if options.skip_resources:
            cmd = cmd.skip_resources(options.skip_resources)
        return cmd



"""Tests for the flux-local `get cluster` command."""

import pytest

from syrupy.assertion import SnapshotAssertion

from . import run_command


@pytest.mark.parametrize(
    ("args"),
    [
        (["--path", "tests/testdata/cluster"]),
        (["--path", "tests/testdata/cluster2"]),
        (["--path", "tests/testdata/cluster3"]),
        (
            [
                "--path",
                "tests/testdata/cluster3",
                "--sources",
                "cluster=tests/testdata/cluster3",
            ]
        ),
        (["--path", "tests/testdata/cluster4"]),
        (["--path", "tests/testdata/cluster5"]),
        (["--path", "tests/testdata/cluster6"]),
        (["--path", "tests/testdata/cluster7"]),
        (["--all-namespaces", "--path", "tests/testdata/cluster/"]),
        (["--path", "tests/testdata/cluster", "-o", "yaml"]),
        (["--path", "tests/testdata/cluster", "-o", "yaml", "--enable-images"]),
        (["--path", "tests/testdata/cluster8", "-o", "yaml"]),
        (["--path", "tests/testdata/cluster8", "-o", "yaml", "--enable-images"]),
        (
            [
                "--path",
                "tests/testdata/cluster8",
                "-o",
                "yaml",
                "--enable-images",
                "--no-skip-secrets",
            ]
        ),
        (
            [
                "--path",
                "tests/testdata/cluster8",
                "-o",
                "yaml",
                "--enable-images",
                "--only-images",
            ]
        ),
        (
            [
                "--path",
                "tests/testdata/cluster9/clusters/dev",
                "--output",
                "yaml",
                "--enable-images",
            ]
        ),
        (["--path", "tests/testdata/cluster", "-o", "json"]),
        (["--path", "tests/testdata/cluster", "-o", "json", "--enable-images"]),
        (
            [
                "--path",
                "tests/testdata/cluster8",
                "-o",
                "json",
                "--enable-images",
                "--only-images",
            ]
        ),
    ],
    ids=[
        "cluster",
        "cluster2",
        "cluster3-no-source",
        "cluster3",
        "cluster4",
        "cluster5",
        "cluster6",
        "cluster7",
        "all-namespaces",
        "yaml-cluster-no-images",
        "yaml-cluster-images",
        "yaml-cluster8-no-images",
        "yaml-cluster8-images",
        "yaml-cluster8-images-allow-secrets",
        "yaml-cluster8-only-images",
        "yaml-cluster9-only-images-with-oci",
        "json-cluster",
        "json-cluster-images",
        "json-cluster-only-images",
    ],
)
async def test_get_cluster(args: list[str], snapshot: SnapshotAssertion) -> None:
    """Test test get cluster commands."""
    result = await run_command(["get", "cluster"] + args)
    assert result == snapshot



"""Tests for helm library."""

from pathlib import Path
from typing import Any, Generator

import pytest
from aiofiles.os import mkdir

from flux_local import kustomize
from flux_local.helm import Helm
from flux_local.manifest import (
    HelmRelease,
    HelmRepository,
    OCIRepository,
)


@pytest.fixture(name="helm_repo_dir")
def helm_repo_dir_fixture() -> Path | None:
    return None


@pytest.fixture(name="oci_repo_dir")
def oci_repo_dir_fixture() -> Path | None:
    return None


@pytest.fixture(name="tmp_config_path")
def tmp_config_path_fixture(tmp_path_factory: Any) -> Generator[Path, None, None]:
    """Fixture for creating a path used for helm config shared across tests."""
    yield tmp_path_factory.mktemp("test_helm")


@pytest.fixture(name="helm_repos")
async def helm_repos_fixture(helm_repo_dir: Path | None) -> list[dict[str, Any]]:
    """Fixture for creating the HelmRepository objects"""
    if not helm_repo_dir:
        return []
    cmd = kustomize.grep("kind=^HelmRepository$", helm_repo_dir)
    return await cmd.objects()


@pytest.fixture(name="oci_repos")
async def oci_repos_fixture(oci_repo_dir: Path) -> list[dict[str, Any]]:
    """Fixture for creating the OCIRepositoriy objects"""
    if not oci_repo_dir:
        return []
    cmd = kustomize.grep("kind=^OCIRepository$", oci_repo_dir)
    return await cmd.objects()


@pytest.fixture(name="helm")
async def helm_fixture(
    tmp_config_path: Path,
    helm_repos: list[dict[str, Any]],
    oci_repos: list[dict[str, Any]],
) -> Helm:
    """Fixture for creating the Helm object."""
    await mkdir(tmp_config_path / "helm")
    await mkdir(tmp_config_path / "cache")
    helm = Helm(
        tmp_config_path / "helm",
        tmp_config_path / "cache",
    )
    helm.add_repos([HelmRepository.parse_doc(repo) for repo in helm_repos])
    helm.add_repos([OCIRepository.parse_doc(repo) for repo in oci_repos])
    return helm


@pytest.fixture(name="helm_releases")
async def helm_releases_fixture(release_dir: Path) -> list[dict[str, Any]]:
    """Fixture for creating the HelmRelease objects."""
    cmd = kustomize.grep("kind=^HelmRelease$", release_dir)
    return await cmd.objects()


async def test_update(helm: Helm) -> None:
    """Test a helm update command."""
    await helm.update()


@pytest.mark.parametrize(
    ("helm_repo_dir", "release_dir"),
    [
        (
            Path("tests/testdata/cluster/infrastructure/configs"),
            Path("tests/testdata/cluster/infrastructure/controllers"),
        ),
    ],
)
async def test_template(helm: Helm, helm_releases: list[dict[str, Any]]) -> None:
    """Test helm template command."""
    await helm.update()

    assert len(helm_releases) == 2
    release = helm_releases[0]
    obj = await helm.template(HelmRelease.parse_doc(release))
    docs = await obj.grep("kind=ServiceAccount").objects()
    names = [doc.get("metadata", {}).get("name") for doc in docs]
    assert names == ["metallb-controller", "metallb-speaker"]


@pytest.mark.parametrize(
    ("oci_repo_dir", "release_dir"),
    [
        (
            Path("tests/testdata/cluster9/apps/podinfo/"),
            Path("tests/testdata/cluster9/apps/podinfo/"),
        ),
    ],
)
async def test_oci_repository(helm: Helm, helm_releases: list[dict[str, Any]]) -> None:
    """Test helm template command."""
    await helm.update()

    assert len(helm_releases) == 1
    release = helm_releases[0]
    obj = await helm.template(HelmRelease.parse_doc(release))
    docs = await obj.grep("kind=Deployment").objects()
    names = [doc.get("metadata", {}).get("name") for doc in docs]
    assert names == ["podinfo"]



"""Tests for the flux-local `build` command."""

import pytest

from syrupy.assertion import SnapshotAssertion

from . import run_command


@pytest.mark.parametrize(
    ("args"),
    [
        (["tests/testdata/cluster/"]),
        (["tests/testdata/cluster2/"]),
        (["tests/testdata/cluster3/"]),
        (["tests/testdata/cluster6/"]),
        (["--enable-helm", "--skip-crds", "tests/testdata/cluster/"]),
        (
            [
                "--enable-helm",
                "--skip-crds",
                "tests/testdata/cluster6/",
                "-a",
                "batch/v1/CronJob",
            ]
        ),
        (["--enable-helm", "--no-skip-secrets", "tests/testdata/cluster8/"]),
        (
            [
                "--enable-helm",
                "--no-skip-secrets",
                "tests/testdata/cluster9/clusters/dev",
            ]
        ),
    ],
    ids=[
        "build",
        "build-cluster2",
        "build-cluster3",
        "build-cluster6",
        "build-helm-cluster",
        "build-helm-cluster6",
        "build-helm-cluster8-valuesFrom",
        "build-helm-cluster9",
    ],
)
async def test_build_all(args: list[str], snapshot: SnapshotAssertion) -> None:
    """Test build commands."""
    result = await run_command(["build", "all"] + args)
    assert result == snapshot


@pytest.mark.parametrize(
    ("args"),
    [
        (["--path=tests/testdata/cluster/"]),
        (["apps", "--path=tests/testdata/cluster/"]),
        (["apps", "--path=tests/testdata/cluster6/"]),
        (["apps", "--path=tests/testdata/cluster8/"]),
        (["apps-stack", "--path=tests/testdata/cluster9/clusters/dev"]),
    ],
    ids=[
        "build-ks",
        "build-ks-single",
        "build-ks-single-cluster6",
        "build-ks-single-cluster8",
        "build-ks-single-cluster9",
    ],
)
async def test_build_ks(args: list[str], snapshot: SnapshotAssertion) -> None:
    """Test build commands."""
    result = await run_command(["build", "ks"] + args)
    assert result == snapshot


@pytest.mark.parametrize(
    ("args"),
    [
        (["--path=tests/testdata/cluster/"]),
        (["weave-gitops", "--path=tests/testdata/cluster/"]),
        (
            [
                "-A",
                "--path",
                "tests/testdata/cluster3",
                "--sources",
                "cluster=tests/testdata/cluster3",
            ]
        ),
        (
            [
                "renovate",
                "-A",
                "--skip-crds",
                "--path",
                "tests/testdata/cluster6/",
                "-a",
                "batch/v1/CronJob",
            ]
        ),
        (["podinfo", "-n", "podinfo", "--path=tests/testdata/cluster8/"]),
        (["podinfo", "-n", "default", "--path=tests/testdata/cluster9/clusters/dev"]),
    ],
    ids=[
        "build-hr",
        "build-hr-single",
        "build-hr-cluster3",
        "build-hr-single-cluster6",
        "build-hr-single-cluster8",
        "build-hr-single-cluster9",
    ],
)
async def test_build_hr(args: list[str], snapshot: SnapshotAssertion) -> None:
    """Test build commands."""
    result = await run_command(["build", "hr"] + args)
    assert result == snapshot



"""Tests for git_repo."""

from collections.abc import Sequence
import io
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from syrupy.assertion import SnapshotAssertion

from flux_local.git_repo import (
    build_manifest,
    ResourceSelector,
    ResourceVisitor,
    Source,
    PathSelector,
    is_allowed_source,
    adjust_ks_path,
)
from flux_local.kustomize import Kustomize, Stash
from flux_local.command import Task, run_piped
from flux_local.manifest import Kustomization
from flux_local.context import trace

TESTDATA = Path("tests/testdata/cluster")
TESTDATA_FULL_PATH = Path.cwd() / TESTDATA


async def test_build_manifest(snapshot: SnapshotAssertion) -> None:
    """Tests for building the manifest."""

    manifest = await build_manifest(TESTDATA)
    assert manifest.compact_dict() == snapshot


async def test_cluster_selector_disabled(snapshot: SnapshotAssertion) -> None:
    """Tests for building the manifest."""

    query = ResourceSelector()
    query.path.path = TESTDATA
    query.cluster.enabled = False

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot


async def test_kustomization_selector_disabled(snapshot: SnapshotAssertion) -> None:
    """Tests for building the manifest."""

    query = ResourceSelector()
    query.path.path = TESTDATA
    query.kustomization.enabled = False

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot


async def test_helm_release_selector_disabled(snapshot: SnapshotAssertion) -> None:
    """Tests for building the manifest with helm releases disabled."""

    query = ResourceSelector()
    query.path.path = TESTDATA
    query.helm_release.enabled = False

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot


async def test_helm_repo_selector_disabled(snapshot: SnapshotAssertion) -> None:
    """Tests for building the manifest with helm repos disabled."""

    query = ResourceSelector()
    query.path.path = TESTDATA
    query.helm_repo.enabled = False

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot


async def test_kustomization_visitor(snapshot: SnapshotAssertion) -> None:
    """Tests for visiting Kustomizations."""

    query = ResourceSelector()
    query.path.path = TESTDATA

    stream = io.StringIO()
    visits: list[tuple[str, str, str]] = []

    async def write(x: Path, y: Any, cmd: Kustomize | None) -> None:
        visits.append((str(x), y.namespace, y.name))
        if cmd:
            stream.write(await cmd.run())

    query.kustomization.visitor = ResourceVisitor(func=write)

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot

    visits.sort()
    assert visits == snapshot

    content = stream.getvalue()
    assert content
    assert "kind: HelmRelease" in content
    assert "name: metallb" in content


async def test_helm_repo_visitor(snapshot: SnapshotAssertion) -> None:
    """Tests for visiting a HelmRepository objects."""

    query = ResourceSelector()
    query.path.path = TESTDATA

    visits: list[tuple[str, str, str]] = []

    async def append(x: Path, y: Any, z: Any) -> None:
        visits.append((str(x), y.namespace, y.name))

    query.helm_repo.visitor = ResourceVisitor(func=append)

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot

    visits.sort()
    assert visits == snapshot


async def test_helm_release_visitor(snapshot: SnapshotAssertion) -> None:
    """Tests for visiting a HelmRelease objects."""

    query = ResourceSelector()
    query.path.path = TESTDATA

    visits: list[tuple[str, str, str]] = []

    async def append(x: Path, y: Any, z: Any) -> None:
        visits.append((str(x), y.namespace, y.name))

    query.helm_release.visitor = ResourceVisitor(
        func=append,
    )

    manifest = await build_manifest(selector=query)
    assert manifest.compact_dict() == snapshot

    visits.sort()
    assert visits == snapshot


def test_source() -> None:
    """Test parsing a source from a string."""
    source = Source.from_str("cluster=./k8s")
    assert source.name == "cluster"
    assert source.namespace is None
    assert str(source.root) == "k8s"


def test_source_with_namespace() -> None:
    """Test parsing a source from a string."""
    source = Source.from_str("flux-system2/cluster=./k8s")
    assert source.name == "cluster"
    assert source.namespace == "flux-system2"
    assert str(source.root) == "k8s"


def test_source_without_path() -> None:
    """Test parsing a source without a path."""
    source = Source.from_str("cluster")
    assert source.name == "cluster"
    assert source.namespace is None
    assert source.root is None


def test_is_allowed_source() -> None:
    """Test GitRepository sources allowed."""
    ks = Kustomization(
        name="ks",
        namespace="flux-system",
        path="./kubernetes/apps/volsync/volsync/app",
        source_name="flux-system",
    )
    assert is_allowed_source([Source.from_str("flux-system")])(ks)


def test_is_not_allowed_source() -> None:
    """Test GitRepository sources allowed."""
    ks = Kustomization(
        name="ks",
        namespace="flux-system",
        path="./kubernetes/apps/volsync/volsync/app",
        source_name="flux-system",
    )
    assert not is_allowed_source([Source.from_str("flux-system-other")])(ks)


def test_is_allowed_source_namespace_optional() -> None:
    """Test GitRepository sources allowed."""
    ks = Kustomization(
        name="ks",
        namespace="flux-system",
        path="./kubernetes/apps/volsync/volsync/app",
        source_name="flux-system",
        source_namespace="flux-system2",
    )
    assert is_allowed_source([Source.from_str("flux-system")])(ks)
    assert is_allowed_source([Source.from_str("flux-system2/flux-system")])(ks)
    assert not is_allowed_source([Source.from_str("flux-system3/flux-system")])(ks)


@pytest.mark.parametrize(
    ("path", "sources"),
    [
        ("tests/testdata/cluster", None),
        ("tests/testdata/cluster2", None),
        (
            "tests/testdata/cluster3",
            [Source.from_str("cluster=tests/testdata/cluster3")],
        ),
        ("tests/testdata/cluster4", None),
        ("tests/testdata/cluster5", None),
        ("tests/testdata/cluster6", None),
        ("tests/testdata/cluster7", None),
    ],
    ids=[
        "cluster",
        "cluster2",
        "cluster3",
        "cluster4",
        "cluster5",
        "cluster6",
        "cluster7",
    ],
)
async def test_internal_commands(
    path: str, sources: list[Source] | None, snapshot: SnapshotAssertion
) -> None:
    """Tests that trace internal command run for each step."""

    context_cmds: dict[str, Any] = {}

    async def piped_recorder(tasks: Sequence[Task]) -> str:
        """Record commands run with the trace context."""

        piped_cmds = []
        for task in tasks:
            if isinstance(task, Stash):
                continue
            piped_cmds.append(str(task))

        c = context_cmds
        for level in trace.get(["root"]):
            if level not in c:
                c[level] = {}
            c = c[level]
        c["cmds"] = c.get("cmds", []) + piped_cmds
        return await run_piped(tasks)

    selector = ResourceSelector(path=PathSelector(path=Path(path), sources=sources))
    with patch("flux_local.kustomize.run_piped", side_effect=piped_recorder):
        await build_manifest(selector=selector)

    assert context_cmds == snapshot


@pytest.mark.parametrize(
    ("path", "expected_path"),
    [
        ("kubernetes/apps/example", "kubernetes/apps/example"),
        ("./kubernetes/apps/example", "kubernetes/apps/example"),
        ("/kubernetes/apps/example", "kubernetes/apps/example"),
        ("", str(TESTDATA)),
    ],
)
def test_adjust_ks_path(path: str, expected_path: str) -> None:
    """Test adjusting the Kustomization path relative directory."""
    ks = Kustomization(
        name="ks",
        namespace="flux-system",
        path=path,
        source_name="flux-system",
    )
    selector = PathSelector(TESTDATA_FULL_PATH)
    assert adjust_ks_path(ks, selector) == Path(expected_path)


async def test_kustomization_label_selector() -> None:
    """Tests for building the manifest."""

    query = ResourceSelector()
    query.path.path = TESTDATA

    async def get_ks() -> list[str]:
        manifest = await build_manifest(selector=query)
        return [
            ks.namespaced_name
            for cluster in manifest.clusters
            for ks in cluster.kustomizations
        ]

    assert await get_ks() == [
        "flux-system/apps",
        "flux-system/flux-system",
        "flux-system/infra-configs",
        "flux-system/infra-controllers",
    ]

    query.kustomization.label_selector = {"app.kubernetes.io/name": "apps"}
    assert await get_ks() == [
        "flux-system/apps",
    ]

    query.kustomization.label_selector = {"app.kubernetes.io/name": "podinfo"}
    assert await get_ks() == []

    # Match on multiple fields
    query.kustomization.label_selector = {
        "app.kubernetes.io/name": "apps",
        "app.kubernetes.io/instance": "apps",
    }
    assert await get_ks() == [
        "flux-system/apps",
    ]

    # Mismatch on one field
    query.kustomization.label_selector = {
        "app.kubernetes.io/name": "apps",
        "app.kubernetes.io/instance": "flux-system",
    }
    assert await get_ks() == []


async def test_helmrelease_label_selector() -> None:
    """Tests for building the manifest."""

    query = ResourceSelector()
    query.path.path = TESTDATA

    async def get_hr() -> list[str]:
        manifest = await build_manifest(selector=query)
        return [
            hr.namespaced_name
            for cluster in manifest.clusters
            for hr in cluster.helm_releases
        ]

    assert await get_hr() == [
        "podinfo/podinfo",
        "metallb/metallb",
        "flux-system/weave-gitops",
    ]

    query.helm_release.label_selector = {"app.kubernetes.io/name": "podinfo"}
    assert await get_hr() == [
        "podinfo/podinfo",
    ]

    query.helm_release.label_selector = {
        "app.kubernetes.io/name": "kubernetes-dashboard"
    }
    assert await get_hr() == []



"""Command line tool for building, diffing, validating flux local repositories."""

import argparse
import asyncio
import logging
import sys
import traceback
from typing import Any

import yaml

from . import build, diff, get, test, diagnostics
from flux_local.exceptions import FluxException

_LOGGER = logging.getLogger(__name__)


def _make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Command line utility for inspecting a local flux repository.",
    )
    parser.add_argument(
        "--log-level", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    )

    subparsers = parser.add_subparsers(dest="command", help="Command", required=True)

    build.BuildAction.register(subparsers)
    get.GetAction.register(subparsers)
    diff.DiffAction.register(subparsers)
    test.TestAction.register(subparsers)
    diagnostics.DiagnosticsAction.register(subparsers)
    return parser


def main() -> None:
    """Flux-local command line tool main entry point."""

    def str_presenter(dumper: yaml.Dumper, data: Any) -> Any:
        """Represent multi-line yaml strings as you'd expect.

        See https://github.com/yaml/pyyaml/issues/240
        """
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", data, style="|" if data.count("\n") > 0 else None
        )

    yaml.add_representer(str, str_presenter)

    # https://github.com/yaml/pyyaml/issues/89
    yaml.Loader.yaml_implicit_resolvers.pop("=")

    parser = _make_parser()
    args = parser.parse_args()

    if args.log_level:
        logging.basicConfig(level=args.log_level)

    action = args.cls()
    try:
        asyncio.run(action.run(**vars(args)))
    except FluxException as err:
        if args.log_level == "DEBUG":
            traceback.print_exc(file=sys.stderr)
        print("flux-local error: ", err, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Utilities for context tracing."""

import contextvars
from contextlib import contextmanager
import logging
from time import perf_counter
from typing import Generator


_LOGGER = logging.getLogger(__name__)

# No public API
__all__: list[str] = []


trace: contextvars.ContextVar[list[str]] = contextvars.ContextVar("trace")


@contextmanager
def trace_context(name: str) -> Generator[None, None, None]:
    stack = trace.get([])
    token = trace.set(stack + [name])
    label = " > ".join(stack + [name])
    t1 = perf_counter()
    _LOGGER.debug("[Trace] > %s", label)
    try:
        yield
    finally:
        t2 = perf_counter()
        trace.reset(token)
        _LOGGER.debug("[Trace] < %s (%0.2fs)", label, (t2 - t1))



"""Module for holding state while visiting resources.

This is used internally primarily for keeping state when building output
or computing diffs.
"""

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass
import logging
import pathlib
import tempfile
import yaml
from typing import Any

from flux_local import git_repo, image
from flux_local.helm import Helm, Options
from flux_local.kustomize import Kustomize
from flux_local.manifest import (
    HelmRelease,
    Kustomization,
    HelmRepository,
    Manifest,
    OCIRepository,
    NamedResource,
)


_LOGGER = logging.getLogger(__name__)


# Strip any annotations from kustomize that contribute to diff noise when
# objects are re-ordered in the output
STRIP_ATTRIBUTES = [
    "config.kubernetes.io/index",
    "internal.config.kubernetes.io/index",
]


ResourceType = Kustomization | HelmRelease | HelmRepository | OCIRepository


@dataclass(frozen=True, order=True)
class ResourceKey:
    """Key for a Kustomization object output."""

    kustomization_path: str
    kind: str
    namespace: str | None
    name: str

    def __post_init__(self) -> None:
        if self.kustomization_path.startswith("/"):
            raise AssertionError(
                f"Expected kustomization_path as relative: {self.kustomization_path}"
            )

    @property
    def label(self) -> str:
        parts = []
        # Either path is a unique identifier within the git repo so prefer the
        # most specific path first.
        if self.kustomization_path and self.kustomization_path != ".":
            parts.append(self.kustomization_path)
            parts.append(" ")
        parts.append(self.compact_label)
        return "".join(parts)

    @property
    def namespaced_name(self) -> str:
        if self.namespace:
            return f"{self.namespace}/{self.name}"
        return self.name

    @property
    def compact_label(self) -> str:
        return f"{self.kind}: {self.namespaced_name}"

    @property
    def named_resource(self) -> NamedResource:
        return NamedResource(
            kind=self.kind,
            name=self.name,
            namespace=self.namespace,
        )


class ResourceOutput(ABC):
    """Helper object for implementing a git_repo.ResourceVisitor that saves content.

    This effectively binds the resource name to the content for later
    inspection by name.
    """

    def visitor(self) -> git_repo.ResourceVisitor:
        """Return a git_repo.ResourceVisitor that points to this object."""
        return git_repo.ResourceVisitor(func=self.call_async)

    @abstractmethod
    async def call_async(
        self,
        kustomization_path: pathlib.Path,
        doc: ResourceType,
        cmd: Kustomize | None,
    ) -> None:
        """Visitor function invoked to record build output."""

    def key_func(
        self,
        kustomization_path: pathlib.Path,
        resource: ResourceType,
    ) -> ResourceKey:
        return ResourceKey(
            kustomization_path=str(kustomization_path),
            kind=resource.__class__.__name__,
            namespace=resource.namespace,
            name=resource.name,
        )


class ContentOutput(ResourceOutput):
    """Resource visitor that build string outputs."""

    def __init__(self) -> None:
        """Initialize KustomizationContentOutput."""
        self.content: dict[ResourceKey, list[str]] = {}

    async def call_async(
        self,
        kustomization_path: pathlib.Path,
        doc: ResourceType,
        cmd: Kustomize | None,
    ) -> None:
        """Visitor function invoked to record build output."""
        if cmd:
            content = await cmd.run()
            lines = content.split("\n")
            if lines[0] != "---":
                lines.insert(0, "---")
            self.content[self.key_func(kustomization_path, doc)] = lines


def strip_attrs(metadata: dict[str, Any], strip_attributes: list[str]) -> None:
    """Update the resource object, stripping any requested labels to simplify diff."""

    for attr_key in ("annotations", "labels"):
        if not (val := metadata.get(attr_key)):
            continue
        for key in strip_attributes:
            if key in val:
                del val[key]
            if not val:
                del metadata[attr_key]
                break


class ImageOutput(ResourceOutput):
    """Resource visitor that builds outputs for objects within the kustomization."""

    def __init__(self) -> None:
        """Initialize ObjectOutput."""
        self.image_visitor = image.ImageVisitor()
        self.repo_visitor = self.image_visitor.repo_visitor()

    async def call_async(
        self,
        kustomization_path: pathlib.Path,
        doc: ResourceType,
        cmd: Kustomize | None,
    ) -> None:
        """Visitor function invoked to build and record resource objects."""
        if cmd and isinstance(doc, HelmRelease):
            objects = await cmd.objects()
            for obj in objects:
                if obj.get("kind") in self.repo_visitor.kinds:
                    self.repo_visitor.func(doc.namespaced_name, obj)

    def update_manifest(self, manifest: Manifest) -> None:
        """Update the manifest with the images found in the repo."""
        for cluster in manifest.clusters:
            for kustomization in cluster.kustomizations:
                for helm_release in kustomization.helm_releases:
                    if images := self.image_visitor.images.get(
                        helm_release.namespaced_name
                    ):
                        helm_release.images = list(images)
                        helm_release.images.sort()


def strip_resource_attributes(
    resource: dict[str, Any], strip_attributes: list[str]
) -> None:
    """Strip any annotations from kustomize that contribute to diff noise when objects are re-ordered in the output."""
    strip_attrs(resource["metadata"], strip_attributes)
    # Remove common noisy labels in commonly used templates
    if (
        (spec := resource.get("spec"))
        and (templ := spec.get("template"))
        and (meta := templ.get("metadata"))
    ):
        strip_attrs(meta, strip_attributes)
    if (
        resource["kind"] == "List"
        and (items := resource.get("items"))
        and isinstance(items, list)
    ):
        for item in items:
            if not (item_meta := item.get("metadata")):
                continue
            strip_attrs(item_meta, strip_attributes)


class ObjectOutput(ResourceOutput):
    """Resource visitor that builds outputs for objects within the kustomization."""

    def __init__(self, strip_attributes: list[str] | None) -> None:
        """Initialize ObjectOutput."""
        # Map of kustomizations to the map of built objects as lines of the yaml string
        self.content: dict[ResourceKey, dict[ResourceKey, list[str]]] = {}
        self.strip_attributes = STRIP_ATTRIBUTES
        if strip_attributes:
            self.strip_attributes.extend(strip_attributes)

    async def call_async(
        self,
        kustomization_path: pathlib.Path,
        doc: ResourceType,
        cmd: Kustomize | None,
    ) -> None:
        """Visitor function invoked to build and record resource objects."""
        if cmd:
            contents: dict[ResourceKey, list[str]] = {}
            objects = await cmd.objects()
            for resource in objects:
                if not (kind := resource.get("kind")) or not (
                    metadata := resource.get("metadata")
                ):
                    _LOGGER.warning(
                        "Invalid document did not contain kind or metadata: %s",
                        resource,
                    )
                    continue
                # Remove common noisy labels
                strip_resource_attributes(resource, self.strip_attributes)

                resource_key = ResourceKey(
                    kind=kind,
                    kustomization_path=str(kustomization_path),
                    namespace=metadata.get("namespace", doc.namespace),
                    name=metadata.get("name", ""),
                )
                content = yaml.dump(resource, sort_keys=False)
                lines = content.split("\n")
                lines.insert(0, "---")
                contents[resource_key] = lines
            self.content[self.key_func(kustomization_path, doc)] = contents


async def inflate_release(
    helm: Helm,
    release: HelmRelease,
    visitor: git_repo.ResourceVisitor,
    options: Options,
) -> None:
    cmd = await helm.template(release, options)
    # We can ignore the Kustomiation path since we're essentially grouping by cluster
    await visitor.func(pathlib.Path(""), release, cmd)


class HelmVisitor:
    """Helper that visits Helm related objects and handles inflation."""

    def __init__(self) -> None:
        """Initialize KustomizationContentOutput."""
        self.repos: list[HelmRepository | OCIRepository] = []
        self.releases: list[HelmRelease] = []

    @property
    def active_repos(self) -> list[HelmRepository | OCIRepository]:
        """Return HelpRepositories referenced by a HelmRelease."""
        repo_keys: set[str] = {
            release.chart.repo_full_name for release in self.releases
        }
        return [repo for repo in self.repos if repo.repo_name in repo_keys]

    def repo_visitor(self) -> git_repo.ResourceVisitor:
        """Return a git_repo.ResourceVisitor that points to this object."""

        async def add_repo(
            kustomization_path: pathlib.Path,
            doc: ResourceType,
            cmd: Kustomize | None,
        ) -> None:
            if not isinstance(doc, HelmRepository) and not isinstance(
                doc, OCIRepository
            ):
                raise ValueError(f"Expected HelmRepository or OCIRepository: {doc}")
            self.repos.append(doc)

        return git_repo.ResourceVisitor(func=add_repo)

    def release_visitor(self) -> git_repo.ResourceVisitor:
        """Return a git_repo.ResourceVisitor that points to this object."""

        async def add_release(
            kustomization_path: pathlib.Path,
            doc: ResourceType,
            cmd: Kustomize | None,
        ) -> None:
            if not isinstance(doc, HelmRelease):
                raise ValueError(f"Expected HelmRelease: {doc}")
            self.releases.append(doc)

        return git_repo.ResourceVisitor(func=add_release)

    async def inflate(
        self,
        helm_cache_dir: pathlib.Path,
        visitor: git_repo.ResourceVisitor,
        options: Options,
    ) -> None:
        """Expand and notify about HelmReleases discovered."""
        _LOGGER.debug("Inflating Helm charts in cluster")
        if not self.releases:
            return
        with tempfile.TemporaryDirectory() as tmp_dir:
            helm = Helm(pathlib.Path(tmp_dir), helm_cache_dir)
            if active_repos := self.active_repos:
                helm.add_repos(active_repos)
                await helm.update()
            tasks = [
                inflate_release(
                    helm,
                    release,
                    visitor,
                    options,
                )
                for release in self.releases
            ]
            _LOGGER.debug("Waiting for inflate tasks to complete")
            await asyncio.gather(*tasks)



"""Flux-local get action."""

import logging
from argparse import (
    ArgumentParser,
    BooleanOptionalAction,
    _SubParsersAction as SubParsersAction,
)
from typing import cast, Any
import sys
import pathlib
import tempfile

from flux_local import git_repo, image, helm
from flux_local.visitor import HelmVisitor, ImageOutput

from .format import (
    PrintFormatter,
    YamlFormatter,
    YamlListFormatter,
    JsonFormatter,
    StructFormatter,
)
from . import selector


_LOGGER = logging.getLogger(__name__)

DEFAULT_NAMESPACE = "flux-system"


class GetKustomizationAction:
    """Get details about kustomizations."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "kustomizations",
                aliases=["ks", "kustomization"],
                help="Get Kustomization objects",
                description="Print information about local flux Kustomization objects",
            ),
        )
        selector.add_ks_selector_flags(args)
        args.add_argument(
            "--output",
            "-o",
            choices=["wide"],
            default=None,
            help="Output format of the command",
        )
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        output: str | None,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        query = selector.build_ks_selector(**kwargs)
        if output != "wide":
            query.helm_release.enabled = False
            query.helm_repo.enabled = False
        manifest = await git_repo.build_manifest(
            selector=query, options=selector.options(**kwargs)
        )

        results: list[dict[str, str]] = []
        cols = ["name", "path"]
        if output == "wide":
            cols.extend(["helmrepos", "ocirepos", "releases"])
        if query.kustomization.namespace is None:
            cols.insert(0, "namespace")
        if len(manifest.clusters) > 1:
            cols.insert(0, "cluster")
        for cluster in manifest.clusters:
            for ks in cluster.kustomizations:
                value = {k: v for k, v in ks.compact_dict().items() if k in cols}
                if output == "wide":
                    value["helmrepos"] = len(ks.helm_repos)
                    value["ocirepos"] = len(ks.oci_repos)
                    value["releases"] = len(ks.helm_releases)
                value["cluster"] = cluster.path
                results.append(value)

        if not results:
            print(selector.not_found("Kustomization", query.kustomization))
            return

        PrintFormatter(cols).print(results)


class GetHelmReleaseAction:
    """Get details about HelmReleases."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "helmreleases",
                aliases=["hr", "helmrelease"],
                help="Get HelmRelease objects",
                description="Print information about local flux HelmRelease objects",
            ),
        )
        selector.add_hr_selector_flags(args)
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        query = selector.build_hr_selector(**kwargs)
        manifest = await git_repo.build_manifest(
            selector=query, options=selector.options(**kwargs)
        )

        cols = ["name", "revision", "chart", "source"]
        if query.helm_release.namespace is None:
            cols.insert(0, "namespace")
        results: list[dict[str, Any]] = []
        for cluster in manifest.clusters:
            for helmrelease in cluster.helm_releases:
                value = {
                    k: v for k, v in helmrelease.compact_dict().items() if k in cols
                }
                value["revision"] = str(helmrelease.chart.version)
                value["chart"] = f"{helmrelease.namespace}-{helmrelease.chart.name}"
                value["source"] = helmrelease.chart.repo_name
                results.append(value)

        if not results:
            print(selector.not_found("HelmRelease", query.helm_release))
            return

        PrintFormatter(cols).print(results)


class GetClusterAction:
    """Get details about flux clustaers."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "clusters",
                aliases=["cl", "cluster"],
                help="Get get flux cluster definitions",
                description="Print information about local flux cluster definitions",
            ),
        )
        selector.add_cluster_selector_flags(args)
        args.add_argument(
            "--enable-images",
            type=str,
            default=False,
            action=BooleanOptionalAction,
            help="Output container images when traversing the cluster",
        )
        args.add_argument(
            "--only-images",
            type=str,
            default=False,
            action=BooleanOptionalAction,
            help="Output only container images when traversing the cluster",
        )
        args.add_argument(
            "--output",
            "-o",
            choices=["diff", "yaml", "json"],
            default="diff",
            help="Output format of the command",
        )
        args.add_argument(
            "--output-file",
            type=str,
            default="/dev/stdout",
            help="Output file for the results of the command",
        )
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        output: str,
        output_file: str,
        enable_images: bool,
        only_images: bool,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        if output not in {"yaml", "json"}:
            if enable_images:
                print(
                    "Flag --enable-images only works with --output yaml or json",
                    file=sys.stderr,
                )
                return
        if only_images and not enable_images:
            print(
                "Flag --only-images only works with --enable-images",
                file=sys.stderr,
            )
            return

        query = selector.build_cluster_selector(**kwargs)
        query.helm_release.enabled = output in {"yaml", "json"}

        image_visitor: image.ImageVisitor | None = None
        helm_content: ImageOutput | None = None
        if enable_images:
            image_visitor = image.ImageVisitor()
            query.doc_visitor = image_visitor.repo_visitor()

            helm_content = ImageOutput()
            helm_visitor = HelmVisitor()
            query.helm_repo.visitor = helm_visitor.repo_visitor()
            query.oci_repo.visitor = helm_visitor.repo_visitor()
            query.helm_release.visitor = helm_visitor.release_visitor()

        manifest = await git_repo.build_manifest(
            selector=query, options=selector.options(**kwargs)
        )
        if output == "yaml" or output == "json":
            if image_visitor:
                image_visitor.update_manifest(manifest)
            if helm_content:
                with tempfile.TemporaryDirectory() as helm_cache_dir:
                    await helm_visitor.inflate(
                        pathlib.Path(helm_cache_dir),
                        helm_content.visitor(),
                        helm.Options(),
                    )
                    helm_content.update_manifest(manifest)

            output_content: Any
            formatter: StructFormatter
            if only_images:
                output_content = sorted(
                    list(
                        {
                            *[
                                image
                                for cluster in manifest.clusters
                                for hr in cluster.helm_releases
                                for image in hr.images or ()
                            ],
                            *[
                                image
                                for cluster in manifest.clusters
                                for ks in cluster.kustomizations
                                for image in ks.images or ()
                            ],
                        }
                    )
                )
                formatter = YamlListFormatter() if output == "yaml" else JsonFormatter()
            else:
                output_content = manifest.compact_dict()
                if output == "yaml":  # Yaml is printing multiple docs
                    output_content = [output_content]
                formatter = YamlFormatter() if output == "yaml" else JsonFormatter()
            with open(output_file, "w") as file:
                formatter.print(output_content, file=file)
            return

        cols = ["path", "kustomizations"]
        results: list[dict[str, Any]] = []
        for cluster in manifest.clusters:
            value = {k: v for k, v in cluster.compact_dict().items() if k in cols}
            value["kustomizations"] = len(cluster.kustomizations)
            results.append(value)

        with open(output_file, "w") as file:
            if not results:
                print(
                    selector.not_found("flux cluster Kustmization", query.cluster),
                    file=file,
                )
                return

            PrintFormatter(cols).print(results, file=file)


class GetAction:
    """Flux-local get action."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "get",
                help="Print information about local flux resources",
                description="Print information about supported local flux resources",
            ),
        )
        subcmds = args.add_subparsers(
            title="Available commands",
            required=True,
        )
        GetKustomizationAction.register(subcmds)
        GetHelmReleaseAction.register(subcmds)
        GetClusterAction.register(subcmds)
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        # No-op given subcommands are always the dispatch target



"""Tests for the format library."""

from flux_local.tool.format import format_columns, PrintFormatter, YamlFormatter


def test_format_columns_empty() -> None:
    """Tests with no rows."""
    assert list(format_columns([], [])) == []


def test_format_columns_empty_rows() -> None:
    """Tests with no rows."""
    assert list(format_columns(["a", "b", "c"], [])) == ["a    b    c    "]


def test_format_columns_rws() -> None:
    """Tests format with normal rows"""
    assert list(
        format_columns(
            ["name", "namespace"], [["podinfo", "podinfo"], ["metallb", "network"]]
        )
    ) == [
        "name       namespace    ",
        "podinfo    podinfo      ",
        "metallb    network      ",
    ]


def test_print_formatter() -> None:
    """Print formatting with empty data."""
    formatter = PrintFormatter()
    assert list(formatter.format([])) == []


def test_print_formatter_data() -> None:
    """Print formatting data objects."""
    formatter = PrintFormatter()
    assert list(
        formatter.format(
            [
                {
                    "name": "podinfo",
                    "namespace": "podinfo",
                },
                {
                    "name": "metallb",
                    "namespace": "network",
                },
            ]
        )
    ) == [
        "NAME       NAMESPACE    ",
        "podinfo    podinfo      ",
        "metallb    network      ",
    ]


def test_print_formatter_keys() -> None:
    """Print formatting with column names."""
    formatter = PrintFormatter(keys=["name"])
    assert list(
        formatter.format(
            [
                {
                    "name": "podinfo",
                    "namespace": "podinfo",
                },
                {
                    "name": "metallb",
                    "namespace": "network",
                },
            ],
        )
    ) == [
        "NAME       ",
        "podinfo    ",
        "metallb    ",
    ]


def test_yaml_formatter() -> None:
    """Print formatting with column names."""
    formatter = YamlFormatter()
    assert list(
        formatter.format(
            [
                {
                    "kustomizations": [
                        {
                            "name": "podinfo",
                            "namespace": "podinfo",
                        },
                        {
                            "name": "metallb",
                            "namespace": "network",
                        },
                    ],
                }
            ]
        )
    ) == [
        "---",
        "kustomizations:",
        "- name: podinfo",
        "  namespace: podinfo",
        "- name: metallb",
        "  namespace: network",
        "",
    ]



"""Tests for helm library."""

import base64
from pathlib import Path
import yaml
from typing import Any

import pytest
from syrupy.assertion import SnapshotAssertion

from flux_local.exceptions import HelmException
from flux_local.values import (
    expand_value_references,
    expand_postbuild_substitute_reference,
    ks_cluster_config,
    cluster_config,
)
from flux_local.manifest import (
    HelmRelease,
    ValuesReference,
    Secret,
    ConfigMap,
    Kustomization,
    HelmChart,
    SubstituteReference,
)
from flux_local.git_repo import ResourceSelector, PathSelector, build_manifest


async def test_value_references(snapshot: SnapshotAssertion) -> None:
    """Test for expanding value references."""
    path = Path("tests/testdata/cluster8")
    selector = ResourceSelector(path=PathSelector(path=path))
    manifest = await build_manifest(selector=selector)
    assert len(manifest.clusters) == 1
    assert len(manifest.clusters[0].kustomizations) == 2
    ks = manifest.clusters[0].kustomizations[0]
    assert ks.name == "apps"
    assert len(ks.helm_releases) == 2
    hr = ks.helm_releases[0]
    assert hr.name == "podinfo"
    assert hr.values == snapshot
    hr = ks.helm_releases[1]
    assert hr.name == "tailscale-operator"
    assert hr.values == snapshot


def test_values_references_with_values_key() -> None:
    """Test for expanding a value reference with a values key."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
                values_key="some-key",
                target_path="target.path",
            ),
            ValuesReference(
                kind="ConfigMap",
                name="test-binary-data-configmap",
                values_key="some-key",
            ),
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-values-configmap",
                namespace="test",
                data={"some-key": "example_value"},
            ),
            ConfigMap(
                name="test-binary-data-configmap",
                namespace="test",
                binary_data={
                    "some-key": base64.b64encode(
                        "encoded_key: encoded_value".encode("utf-8")
                    )
                },
            ),
        ],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == {
        "test": "test",
        "target": {
            "path": "example_value",
        },
        "encoded_key": "encoded_value",
    }



@pytest.mark.parametrize(
    ("inline_values", "expected_values"),
    [
        (
            {
                "redis": {
                    "enabled": True,
                }
            },
            {
                "redis": {
                    "enabled": True,
                }
            }
        ),
        (
            {},
            {
                "redis": {
                    "enabled": False,
                }
            }
        )
    ]
)
def test_value_reference_ordering(inline_values: dict[str, Any], expected_values: dict[str, Any]) -> None:
    """Test for expanding a value reference with inline values that overwrite."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={
            **inline_values,
        },
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-binary-data-configmap",
                values_key="some-key",
            ),
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-binary-data-configmap",
                namespace="test",
                binary_data={
                    "some-key": base64.b64encode(
                        "redis:\n  enabled: False".encode("utf-8")
                    )
                },
            ),
        ],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == expected_values


def test_values_references_with_missing_values_key() -> None:
    """Test for expanding a value reference with a values key that is missing."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
                values_key="some-key-does-not-exist",
                target_path="target.path",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-values-configmap",
                namespace="test",
                data={"some-key": "example_value"},
            )
        ],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == {
        "test": "test",
    }


def test_values_references_with_missing_secret() -> None:
    """Test for expanding a value reference with a missing secret."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="Secret",
                name="test-values-secret",
                values_key="some-key-does-not-exist",
                target_path="target.path",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == {
        "test": "test",
        "target": {
            "path": "..PLACEHOLDER_test-values-secret..",
        },
    }


def test_values_references_with_missing_secret_values_key() -> None:
    """Test for expanding a value reference with a secret key that is missing."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="Secret",
                name="test-values-secret",
                values_key="some-key-does-not-exist",
                target_path="target.path",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        secrets=[
            Secret(
                name="test-values-secret",
                namespace="test",
                string_data={"some-key": "example_value"},
            )
        ],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == {
        "test": "test",
    }


def test_values_references_invalid_yaml() -> None:
    """Test for expanding a value reference with a values key."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-values-configmap",
                namespace="test",
                data={"values.yaml": "not-yaml"},
            )
        ],
    )
    with pytest.raises(HelmException, match=r"valid yaml"):
        expand_value_references(hr, ks)


def test_values_references_invalid_binary_data() -> None:
    """Test for expanding a value reference with an invalid binary data key."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-values-configmap",
                namespace="test",
                binary_data={"values.yaml": "this is not base64 data"},
            )
        ],
    )
    with pytest.raises(HelmException, match=r"Unable to decode binary data"):
        expand_value_references(hr, ks)


def test_values_reference_invalid_target_path() -> None:
    """Test for expanding a value reference with a values key."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-original-configmap",
            ),
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
                values_key="some-key",
                # Target above is a list
                target_path="target.path",
            )
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[
            ConfigMap(
                name="test-values-configmap",
                namespace="test",
                data={"some-key": "example_value"},
            ),
            ConfigMap(
                name="test-values-original-configmap",
                namespace="test",
                data={
                    "values.yaml": yaml.dump({
                        "test": "test",
                        "target": ["a", "b", "c"],
                    }),
                }
            )
        ],
    )
    with pytest.raises(HelmException, match=r"values to be a dict"):
        expand_value_references(hr, ks)


def test_values_reference_invalid_configmap_and_secret() -> None:
    """Test a values reference to a config map and secret that do not exist."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="ConfigMap",
                name="test-values-configmap",
                optional=False,  # We just log
            ),
            ValuesReference(
                kind="Secret",
                name="test-values-secret",
                optional=False,  # We just log
            ),
            ValuesReference(
                kind="UnknownKind",
                name="test-values-secret",
            ),
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        config_maps=[],
    )
    updated_hr = expand_value_references(hr, ks)
    # No changes to the values
    assert updated_hr.values == {"test": "test"}


def test_values_references_secret() -> None:
    """Test for expanding a value reference for a secret."""
    hr = HelmRelease(
        name="test",
        namespace="test",
        chart=HelmChart(
            repo_name="test-repo",
            repo_namespace="flux-system",
            name="test-chart",
            version="test-version",
        ),
        values={"test": "test"},
        values_from=[
            ValuesReference(
                kind="Secret",
                name="test-values-secret",
                values_key="some-key1",
                target_path="target.path1",
            ),
            ValuesReference(
                kind="Secret",
                name="test-string-values-secret",
                values_key="some-key2",
                target_path="target.path2",
            ),
        ],
    )
    ks = Kustomization(
        name="test",
        namespace="test",
        path="example/path",
        helm_releases=[hr],
        secrets=[
            Secret.parse_doc(
                {
                    "apiVersion": "v1",
                    "kind": "Secret",
                    "metadata": {
                        "name": "test-values-secret",
                        "namespace": "test",
                    },
                    "data": {
                        "some-key1": base64.b64encode("example-value".encode("utf-8")),
                    },
                }
            ),
            Secret.parse_doc(
                {
                    "apiVersion": "v1",
                    "kind": "Secret",
                    "metadata": {
                        "name": "test-string-values-secret",
                        "namespace": "test",
                    },
                    "stringData": {
                        "some-key2": "example-string-value",
                    },
                }
            ),
        ],
    )
    updated_hr = expand_value_references(hr, ks)
    assert updated_hr.values == {
        "test": "test",
        "target": {
            "path1": "..PLACEHOLDER_some-key1..",
            "path2": "..PLACEHOLDER_some-key2..",
        },
    }


def test_postbuild_substitute_from() -> None:
    """Test for expanding from a config map."""
    app_ks = Kustomization(
        name="app",
        namespace="test",
        path="example/app",
        postbuild_substitute={"test": "test"},
        postbuild_substitute_from=[
            SubstituteReference(
                kind="ConfigMap",
                name="test-values-configmap",
            ),
            SubstituteReference(
                kind="ConfigMap",
                name="test-binary-data-configmap",
            ),
        ],
    )
    config_maps = [
        ConfigMap(
            name="test-values-configmap",
            namespace="test",
            data={"some_key": "example_value"},
        ),
        ConfigMap(
            name="test-binary-data-configmap",
            namespace="test",
            binary_data={
                "encoded_key": base64.b64encode("encoded_value".encode("utf-8"))
            },
        ),
    ]
    app_ks = expand_postbuild_substitute_reference(
        app_ks, cluster_config([], config_maps)
    )
    assert app_ks.postbuild_substitute == {
        "test": "test",
        "some_key": "example_value",
        "encoded_key": "encoded_value",
    }


def test_postbuild_substitute_from_secret() -> None:
    """Test for expanding from a secret."""
    app_ks = Kustomization(
        name="app",
        namespace="test",
        path="example/app",
        postbuild_substitute={"test": "test"},
        postbuild_substitute_from=[
            SubstituteReference(
                kind="Secret",
                name="test-values-secret",
            ),
        ],
    )
    secrets = [
        Secret(
            name="test-values-secret",
            namespace="test",
            string_data={"some_key": "example_value"},
        ),
    ]
    app_ks = expand_postbuild_substitute_reference(app_ks, cluster_config(secrets, []))
    assert app_ks.postbuild_substitute == {
        "test": "test",
        "some_key": "example_value",
    }


def test_postbuild_substitute_from_invalid_references() -> None:
    """Test for invalid SubstituteReferences."""
    app_ks = Kustomization(
        name="app",
        namespace="test",
        path="example/app",
        postbuild_substitute={"test": "test"},
        postbuild_substitute_from=[
            SubstituteReference(
                kind="Secret",
                name="test-values-secret",
            ),
            SubstituteReference(
                kind="ConfigMap",
                name="test-values-config-map",
            ),
            SubstituteReference(
                kind="Invalid",
                name="test-values-invalid",
            ),
        ],
    )
    app_ks = expand_postbuild_substitute_reference(app_ks, cluster_config([], []))
    assert app_ks.postbuild_substitute == {
        "test": "test",
    }


def test_cluster_config() -> None:
    """Test for ClusterConfig objects."""

    ks1 = Kustomization(
        name="app",
        namespace="test",
        path="example/app",
        secrets=[
            Secret(
                name="test-values-secret",
                namespace="test",
                string_data={"some_key": "example_value"},
            ),
        ],
    )
    ks2 = Kustomization(
        name="app",
        namespace="test",
        path="example/app",
        secrets=[
            Secret(
                name="other-values-secret",
                namespace="test",
                string_data={"some_key": "example_value"},
            ),
        ],
    )
    config = ks_cluster_config([ks1, ks2])
    # Ensure can be called repeatedly
    assert len(list(config.secrets)) == 2
    assert [s.name for s in config.secrets] == [
        "test-values-secret",
        "other-values-secret",
    ]



"""Flux local test action."""

from argparse import (
    ArgumentParser,
    BooleanOptionalAction,
    _SubParsersAction as SubParsersAction,
)
import asyncio
from dataclasses import dataclass
import logging
import pathlib
import tempfile
from pathlib import Path
import sys
from typing import cast, Generator, Any

import nest_asyncio
import pytest

from flux_local import git_repo, kustomize
from flux_local.exceptions import FluxException
from flux_local.helm import Helm, Options
from flux_local.manifest import (
    Manifest,
    Cluster,
    Kustomization,
    HelmRelease,
    HelmRepository,
    OCIRepository,
)
from . import selector

_LOGGER = logging.getLogger(__name__)


@dataclass
class TestConfig:
    """Test configuration, which are parameters to types of the tests."""

    options: git_repo.Options
    helm_options: Options


class HelmReleaseTest(pytest.Item):
    """Test case for a Kustomization."""

    cluster: Cluster
    kustomization: Kustomization
    helm_release: HelmRelease
    test_config: TestConfig

    @classmethod
    def from_parent(  # type: ignore[override]
        cls,
        parent: Any,
        *,
        cluster: Cluster,
        kustomization: Kustomization,
        helm_release: HelmRelease,
        test_config: TestConfig,
        **kw: Any,
    ) -> "HelmReleaseTest":
        """The public constructor."""
        item: HelmReleaseTest = super().from_parent(
            parent=parent,
            path=Path(kustomization.path),
            name=f"{helm_release.namespace}/{helm_release.name}",
            **kw,
        )
        item.cluster = cluster
        item.kustomization = kustomization
        item.helm_release = helm_release
        item.test_config = test_config
        return item

    def runtest(self) -> None:
        """Dispatch the async work and run the test."""
        nest_asyncio.apply()
        asyncio.run(self.async_runtest())

    async def async_runtest(self) -> None:
        """Run the Kustomizations test."""
        # Note: This could be sped up by sharing a cache dir across clusters for the
        # multi-cluster git repos.
        with (
            tempfile.TemporaryDirectory() as helm_cache_dir,
            tempfile.TemporaryDirectory() as tmp_dir,
        ):
            helm = Helm(pathlib.Path(tmp_dir), pathlib.Path(helm_cache_dir))
            helm.add_repos(self.active_repos())
            await helm.update()
            cmd = await helm.template(
                self.helm_release,
                self.test_config.helm_options,
            )
            await cmd.objects()

    def active_repos(self) -> list[HelmRepository | OCIRepository]:
        """Return HelmRepositories referenced by a HelmRelease."""
        repo_key = "-".join(
            [
                self.helm_release.chart.repo_namespace,
                self.helm_release.chart.repo_name,
            ]
        )
        return [
            repo
            for repo in self.cluster.helm_repos + self.cluster.oci_repos
            if repo.repo_name == repo_key
        ]


class KustomizationTest(pytest.Item):
    """Test case for a Kustomization."""

    cluster: Cluster
    kustomization: Kustomization
    test_config: TestConfig

    @classmethod
    def from_parent(  # type: ignore[override]
        cls,
        parent: Any,
        *,
        cluster: Cluster,
        kustomization: Kustomization,
        test_config: TestConfig,
        **kw: Any,
    ) -> "KustomizationTest":
        """The public constructor."""
        item: KustomizationTest = super().from_parent(
            parent=parent, path=Path(kustomization.path), name="kustomization", **kw
        )
        item.cluster = cluster
        item.kustomization = kustomization
        item.test_config = test_config
        return item

    def runtest(self) -> None:
        """Dispatch the async work and run the test."""
        nest_asyncio.apply()
        asyncio.run(self.async_runtest())

    async def async_runtest(self) -> None:
        """Run the Kustomizations test."""
        cmd = await kustomize.flux_build(
            self.kustomization, Path(self.kustomization.path)
        ).stash()
        await cmd.objects()


class KustomizationCollector(pytest.Collector):
    """Test collector for a Kustomization."""

    cluster: Cluster
    kustomization: Kustomization
    test_config: TestConfig

    @classmethod
    def from_parent(  # type: ignore[override]
        cls,
        parent: Any,
        *,
        cluster: Cluster,
        kustomization: Kustomization,
        test_config: TestConfig,
        **kw: Any,
    ) -> "KustomizationCollector":
        """The public constructor."""
        item: KustomizationCollector = super().from_parent(
            parent=parent, name=kustomization.name, path=Path(kustomization.path), **kw
        )
        item.cluster = cluster
        item.kustomization = kustomization
        item.test_config = test_config
        return item

    def collect(self) -> Generator[pytest.Item | pytest.Collector, None, None]:
        """Collect tests from the cluster manifest.yaml file."""
        yield KustomizationTest.from_parent(
            parent=self,
            cluster=self.cluster,
            kustomization=self.kustomization,
            test_config=self.test_config,
        )
        for helm_release in self.kustomization.helm_releases:
            yield HelmReleaseTest.from_parent(
                parent=self,
                cluster=self.cluster,
                kustomization=self.kustomization,
                helm_release=helm_release,
                test_config=self.test_config,
            )


class ClusterCollector(pytest.Collector):
    """Test collector for a Cluster."""

    cluster: Cluster
    test_config: TestConfig

    @classmethod
    def from_parent(  # type: ignore[override]
        cls, parent: Any, *, cluster: Cluster, test_config: TestConfig, **kw: Any
    ) -> "ClusterCollector":
        """The public constructor."""
        item: ClusterCollector = super().from_parent(
            parent=parent,
            name=cluster.path,
            path=Path(cluster.path),
            nodeid=str(Path(cluster.path)),
        )
        item.cluster = cluster
        item.test_config = test_config
        return item

    def collect(self) -> Generator[pytest.Item | pytest.Collector, None, None]:
        """Collect tests from the cluster."""
        for kustomization in self.cluster.kustomizations:
            yield KustomizationCollector.from_parent(
                parent=self,
                cluster=self.cluster,
                kustomization=kustomization,
                test_config=self.test_config,
            )


class ManifestCollector(pytest.Collector):
    """Test collector for a Kustomization."""

    manifest: Manifest
    test_config: TestConfig

    @classmethod
    def from_parent(  # type: ignore[override]
        cls,
        parent: Any,
        *,
        manifest: Manifest,
        test_config: TestConfig,
        **kw: Any,
    ) -> "ManifestCollector":
        """The public constructor."""
        item: ManifestCollector = super().from_parent(
            parent=parent, name="manifest", **kw
        )
        item.manifest = manifest
        item.test_config = test_config
        return item

    def collect(self) -> Generator[pytest.Item | pytest.Collector, None, None]:
        """Collect tests from the cluster manifest."""
        for cluster in self.manifest.clusters:
            yield ClusterCollector.from_parent(
                parent=self,
                cluster=cluster,
                test_config=self.test_config,
            )


class ManifestPlugin:
    """Loads the flux-local Manifest based on command line arguments.

    This plugin prepares the `test_kustomization` plugin with the manifest. That plugin
    is loaded separately to avoid collection.
    """

    def __init__(
        self,
        selector: git_repo.ResourceSelector,
        test_config: TestConfig,
        test_filter: list[str],
    ) -> None:
        self.selector = selector
        self.manifest: Manifest | None = None
        self.test_config = test_config
        self.test_filter = test_filter
        self.init_error: Exception | None = None

    def pytest_sessionstart(self, session: pytest.Session) -> None:
        nest_asyncio.apply()
        asyncio.run(self.async_pytest_sessionstart(session))

    async def async_pytest_sessionstart(self, session: pytest.Session) -> None:
        """Run the Kustomizations test."""
        _LOGGER.debug("async_pytest_sessionstart")
        try:
            manifest = await git_repo.build_manifest(
                selector=self.selector,
                options=self.test_config.options,
            )
        except FluxException as err:
            _LOGGER.error("Failed to build manifest: %s", err)
            self.init_error = err
            return

        self.manifest = manifest
        _LOGGER.debug("async_pytest_sessionstart ended")

    def pytest_collection(self, session: pytest.Session) -> None:
        _LOGGER.debug("pytest_collection:%s", session)
        if self.init_error or self.manifest is None:
            raise pytest.UsageError(
                self.init_error or "ManifestPlugin not initialized properly"
            ) from self.init_error
        manifest_collector = ManifestCollector.from_parent(
            parent=session,
            manifest=self.manifest,
            test_config=self.test_config,
        )
        # Ignore the default files found by pytest and instead create
        # tests based on the manifest contents.
        session.collect = manifest_collector.collect  # type: ignore[method-assign]
        _LOGGER.debug("pytest_collection end:%s", session)

    def pytest_collection_modifyitems(
        self,
        session: pytest.Session,
        config: pytest.Config,
        items: list[pytest.Item],
    ) -> None:
        _LOGGER.debug("pytest_collection_modifyitems collected: %s", len(items))
        if self.test_filter:
            _LOGGER.debug("Filtering tests: %s", self.test_filter)
            filtered_items = []
            for item in items:
                for nodeid in self.test_filter:
                    if item.nodeid.startswith(nodeid):
                        filtered_items.append(item)
                        continue
            items.clear()
            items.extend(filtered_items)
            _LOGGER.debug("Remaining tests after collection filter: %s", len(items))


class TestAction:
    """Flux-local test action."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "test",
                help="Build and validate the cluster",
                description="""You can verify that the resources in the cluster
                    are formatted properly before commit or as part of a CI
                    system. The flux-local test command will build the
                    Kustomization resources in the cluster.""",
            ),
        )
        args.add_argument(
            "--enable-helm",
            type=bool,
            action=BooleanOptionalAction,
            help="Enable use of HelmRelease inflation",
        )
        # Flags consistent with pytest for pass through
        args.add_argument(
            "test_path",
            help="Optional path with flux Kustomization resources or full test node",
            type=str,
            default=None,
            nargs="?",
        )
        verbosity_group = args.add_mutually_exclusive_group()
        verbosity_group.add_argument(
            "--verbose",
            "-v",
            action="count",
            dest="verbosity",
            help="Increase verbosity.",
        )
        verbosity_group.add_argument(
            "--verbosity",
            action="store",
            type=int,
            metavar="N",
            dest="verbosity",
            help="Set verbosity. Default is 0",
        )
        selector.add_helm_options_flags(args)
        args.set_defaults(cls=cls, verbosity=0)
        selector.add_cluster_selector_flags(args)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        enable_helm: bool,
        test_path: str | None,
        verbosity: int,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        query = selector.build_cluster_selector(**kwargs)
        if test_path:
            parts = test_path.split("::")
            query.path.path = Path(parts[0])

            # If a real file path, then clear so it is not a test nodeid filter
            if test_path.startswith(".") or test_path.startswith("/"):
                test_path = None
        query.kustomization.namespace = query.cluster.namespace
        query.kustomization.skip_crds = True
        query.helm_release.enabled = enable_helm
        query.helm_release.namespace = None
        options = selector.options(**kwargs)
        helm_options = selector.build_helm_options(**kwargs)

        nest_asyncio.apply()
        pytest_args = [
            "--verbosity",
            str(verbosity),
            "--no-header",
            "--disable-warnings",
        ]
        _LOGGER.debug("pytest.main: %s", pytest_args)
        retcode = pytest.main(
            pytest_args,
            plugins=[
                ManifestPlugin(
                    query,
                    TestConfig(
                        options=options,
                        helm_options=helm_options,
                    ),
                    test_filter=[str(test_path)] if test_path else [],
                )
            ],
        )
        if retcode:
            sys.exit(retcode)



"""Tests for the flux-local `get ks` command."""

import pytest

from syrupy.assertion import SnapshotAssertion

from . import run_command


@pytest.mark.parametrize(
    ("args"),
    [
        (["--path", "tests/testdata/cluster"]),
        (["--path", "tests/testdata/cluster2"]),
        (
            [
                "-A",
                "--path",
                "tests/testdata/cluster3",
                "--sources",
                "cluster=tests/testdata/cluster3",
            ]
        ),
        (["--path", "tests/testdata/cluster4"]),
        (["--path", "tests/testdata/cluster5"]),
        (["--all", "--path", "tests/testdata/cluster5"]),
        (["--path", "tests/testdata/cluster6"]),
        (["--path", "tests/testdata/cluster7"]),
        (["--path", "./tests/testdata/cluster/clusters/prod"]),
        (["--path", "tests/testdata/cluster", "-o", "wide"]),
        (["--path", "tests/testdata/cluster2", "-o", "wide"]),
        (["--all-namespaces", "--path", "./tests/testdata/cluster/apps/prod"]),
        (["--path", "tests/testdata/cluster9/clusters/dev"]),
    ],
    ids=[
        "cluster",
        "cluster2",
        "cluster3",
        "cluster4",
        "cluster5",
        "cluster5-all",
        "cluster6",
        "cluster7",
        "cluster_path",
        "wide",
        "cluster2-wide",
        "ks_path",
        "cluster9",
    ],
)
async def test_get_ks(args: list[str], snapshot: SnapshotAssertion) -> None:
    """Test test get ks commands."""
    result = await run_command(["get", "ks"] + args)
    assert result == snapshot



"""Tests for the flux-local `diagnostics` command."""

import pathlib
import tempfile

from syrupy.assertion import SnapshotAssertion
import yaml
import pytest

from . import run_command


async def test_diagnostics(snapshot: SnapshotAssertion) -> None:
    """Test test get ks commands."""
    result = await run_command(["diagnostics"])
    assert result == snapshot


@pytest.mark.parametrize(
    ("contents"),
    [
        (yaml.dump_all([["entry"]]).encode()),
        (yaml.dump_all(["entry", "var"]).encode()),
        (b"\x00\x00\x00"),
    ],
    ids=["list", "scalar", "not-yaml"],
)
async def test_invalid_mapping(contents: bytes, snapshot: SnapshotAssertion) -> None:
    """Test test get ks commands."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        example_file = pathlib.Path(tmp_dir) / "example.yaml"
        example_file.write_bytes(contents)

        result = await run_command(["diagnostics", "--path", tmp_dir])
        result = result.replace(str(example_file), "<<TEST_FILE>>")
        assert result == snapshot
